# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core import (
    Var,
    Block,
    Constraint,
    Param,
    Set,
    SetOf,
    Suffix,
    Expression,
    Objective,
    SortComponents,
    value,
    ConstraintList,
)
from pyomo.core.base import TransformationFactory, VarData
from pyomo.core.plugins.transform.hierarchy import Transformation
from pyomo.common.config import ConfigBlock, ConfigValue, NonNegativeFloat
from pyomo.common.modeling import unique_component_name
from pyomo.repn.standard_repn import generate_standard_repn
from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.opt import TerminationCondition
from pyomo.util.config_domains import ComponentDataSet

import logging

logger = logging.getLogger('pyomo.contrib.fme')


def _check_var_bounds_filter(constraint):
    """Check if the constraint is already implied by the variable bounds"""
    # this is one of our constraints, so we know that it is >=.
    min_lhs = 0
    for v, coef in constraint['map'].items():
        if coef > 0:
            if v.lb is None:
                return True  # we don't have var bounds with which to imply the
                # constraint...
            min_lhs += coef * v.lb
        elif coef < 0:
            if v.ub is None:
                return True  # we don't have var bounds with which to imply the
                # constraint...
            min_lhs += coef * v.ub
    # we do need value here since we didn't control v.lb and v.ub above.
    if value(min_lhs) >= constraint['lower']:
        return False  # constraint implied by var bounds
    return True


def gcd(a, b):
    while b != 0:
        a, b = b, a % b
    return abs(a)


def lcm(ints):
    a = ints[0]
    for b in ints[1:]:
        a = abs(a * b) // gcd(a, b)
    return a


@TransformationFactory.register(
    'contrib.fourier_motzkin_elimination',
    doc="Project out specified (continuous) variables from a linear model.",
)
class Fourier_Motzkin_Elimination_Transformation(Transformation):
    """Project out specified variables from a linear model.

    This transformation requires the following keyword argument:
        vars_to_eliminate: A user-specified list of continuous variables to
                           project out of the model

    The transformation will deactivate the original constraints of the model
    and create a new block named "_pyomo_contrib_fme_transformation" with the
    projected constraints. Note that this transformation will flatten the
    structure of the original model since there is no obvious mapping between
    the original model and the transformed one.

    """

    CONFIG = ConfigBlock("contrib.fourier_motzkin_elimination")
    CONFIG.declare(
        'vars_to_eliminate',
        ConfigValue(
            default=None,
            domain=ComponentDataSet(Var),
            description="Continuous variable or list of continuous variables to "
            "project out of the model",
            doc="""
            This specifies the list of variables to project out of the model.
            Note that these variables must all be continuous and the model must be
            linear.""",
        ),
    )
    CONFIG.declare(
        'constraint_filtering_callback',
        ConfigValue(
            default=_check_var_bounds_filter,
            description="A callback that determines whether or not new "
            "constraints generated by Fourier-Motzkin elimination are added "
            "to the model",
            doc="""
            Specify None in order for no constraint filtering to occur during the
            transformation.
    
            Specify a function that accepts a constraint (represented in the >=
            dictionary form used in this transformation) and returns a Boolean
            indicating whether or not to add it to the model.
            """,
        ),
    )
    CONFIG.declare(
        'do_integer_arithmetic',
        ConfigValue(
            default=False,
            domain=bool,
            description="A Boolean flag to decide whether Fourier-Motzkin "
            "elimination will be performed with only integer arithmetic.",
            doc="""
            If True, only integer arithmetic will be performed during Fourier-
            Motzkin elimination. This should result in no numerical error.
            If True and there is non-integer data in the constraints being
            projected, an error will be raised.
    
            If False, the algorithm will not check whether data is integer, and will
            perform division operations. Use this setting when not all data is
            integer, or when you are willing to sacrifice some numeric accuracy.
            """,
        ),
    )
    CONFIG.declare(
        'verbose',
        ConfigValue(
            default=False,
            domain=bool,
            description="A Boolean flag to enable verbose output.",
            doc="""
            If True, logs the steps of the projection.
            """,
        ),
    )
    CONFIG.declare(
        'zero_tolerance',
        ConfigValue(
            default=0,
            domain=NonNegativeFloat,
            description="Absolute tolerance at which a float will be considered 0.",
            doc="""
            Whenever fourier-motzkin elimination is used with non-integer data,
            there is a chance of numeric trouble, the most obvious of which is
            that 'eliminated' variables will remain in the constraints with very
            small coefficients. Set this tolerance so that floating points smaller
            than this will be treated as 0 (and reported that way in the final
            constraints).
            """,
        ),
    )
    CONFIG.declare(
        'integer_tolerance',
        ConfigValue(
            default=0,
            domain=NonNegativeFloat,
            description="Absolute tolerance at which a float will be considered "
            "(and cast to) an integer, when do_integer_arithmetic is True",
            doc="""
            Tolerance at which a number x will be considered an integer, when we
            are performing fourier-motzkin elimination with only integer_arithmetic.
            That is, x will be cast to an integer if
            abs(int(x) - x) <= integer_tolerance.
            """,
        ),
    )
    CONFIG.declare(
        'projected_constraints_name',
        ConfigValue(
            default=None,
            domain=str,
            description="Optional name for the ConstraintList containing the "
            "projected constraints. Must be a unique name with respect to the "
            "instance.",
            doc="""
            Optional name for the ConstraintList containing the projected 
            constraints. If not specified, the constraints will be stored on a 
            private block created by the transformation, so if you want access 
            to them after the transformation, use this argument.
    
            Must be a string which is a unique component name with respect to the 
            Block on which the transformation is called.
            """,
        ),
    )

    def __init__(self):
        """Initialize transformation object"""
        super(Fourier_Motzkin_Elimination_Transformation, self).__init__()

    def _apply_to(self, instance, **kwds):
        log_level = logger.level
        try:
            config = self.CONFIG(kwds.pop('options', {}))
            config.set_value(kwds)
            # lower logging values emit more
            if config.verbose and log_level > logging.INFO:
                logger.setLevel(logging.INFO)
                self.verbose = True
            # if the user used the logger to ask for info level messages
            elif log_level <= logging.INFO:
                self.verbose = True
            else:
                self.verbose = False
            self._apply_to_impl(instance, config)
        finally:
            # restore logging level
            logger.setLevel(log_level)

    def _apply_to_impl(self, instance, config):
        vars_to_eliminate = config.vars_to_eliminate
        self.constraint_filter = config.constraint_filtering_callback
        self.do_integer_arithmetic = config.do_integer_arithmetic
        self.integer_tolerance = config.integer_tolerance
        self.zero_tolerance = config.zero_tolerance
        if vars_to_eliminate is None:
            raise RuntimeError(
                "The Fourier-Motzkin Elimination transformation "
                "requires the argument vars_to_eliminate, a "
                "list of Vars to be projected out of the model."
            )

        # make transformation block
        transBlockName = unique_component_name(
            instance, '_pyomo_contrib_fme_transformation'
        )
        transBlock = Block()
        instance.add_component(transBlockName, transBlock)
        nm = config.projected_constraints_name
        if nm is None:
            projected_constraints = transBlock.projected_constraints = ConstraintList()
        else:
            # check that this component doesn't already exist
            if instance.component(nm) is not None:
                raise RuntimeError(
                    "projected_constraints_name was specified "
                    "as '%s', but this is already a component "
                    "on the instance! Please specify a unique "
                    "name." % nm
                )
            projected_constraints = ConstraintList()
            instance.add_component(nm, projected_constraints)

        # collect all of the constraints
        # NOTE that we are ignoring deactivated constraints
        constraints = []
        ctypes_not_to_transform = set(
            (Block, Param, Objective, Set, SetOf, Expression, Suffix, Var)
        )
        for obj in instance.component_data_objects(
            descend_into=Block, sort=SortComponents.deterministic, active=True
        ):
            if obj.ctype in ctypes_not_to_transform:
                continue
            elif obj.ctype is Constraint:
                cons_list = self._process_constraint(obj)
                constraints.extend(cons_list)
                obj.deactivate()  # the truth will be on our transformation block
            else:
                raise RuntimeError(
                    "Found active component %s of type %s. The "
                    "Fourier-Motzkin Elimination transformation can only "
                    "handle purely algebraic models. That is, only "
                    "Sets, Params, Vars, Constraints, Expressions, Blocks, "
                    "and Objectives may be active on the model." % (obj.name, obj.ctype)
                )

        for obj in vars_to_eliminate:
            if obj.lb is not None:
                constraints.append(
                    {
                        'body': generate_standard_repn(obj),
                        'lower': value(obj.lb),
                        'map': ComponentMap([(obj, 1)]),
                    }
                )
            if obj.ub is not None:
                constraints.append(
                    {
                        'body': generate_standard_repn(-obj),
                        'lower': -value(obj.ub),
                        'map': ComponentMap([(obj, -1)]),
                    }
                )

        new_constraints = self._fourier_motzkin_elimination(
            constraints, vars_to_eliminate
        )

        # put the new constraints on the transformation block
        for cons in new_constraints:
            if self.constraint_filter is not None:
                try:
                    keep = self.constraint_filter(cons)
                except:
                    logger.error(
                        "Problem calling constraint filter callback "
                        "on constraint with right-hand side %s and "
                        "body:\n%s" % (cons['lower'], cons['body'].to_expression())
                    )
                    raise
                if not keep:
                    continue
            lhs = cons['body'].to_expression(sort=True)
            lower = cons['lower']
            assert type(lower) is int or type(lower) is float
            if type(lhs >= lower) is bool:
                if lhs >= lower:
                    continue
                else:
                    # This would actually make a lot of sense in this case...
                    # projected_constraints.add(Constraint.Infeasible)
                    raise RuntimeError("Fourier-Motzkin found the model is infeasible!")
            else:
                projected_constraints.add(lhs >= lower)

    def _process_constraint(self, constraint):
        """Transforms a pyomo Constraint object into a list of dictionaries
        representing only >= constraints. That is, if the constraint has both an
        ub and a lb, it is transformed into two constraints. Otherwise it is
        flipped if it is <=. Each dictionary contains the keys 'lower',
        and 'body' where, after the process, 'lower' will be a constant, and
        'body' will be the standard repn of the body. (The constant will be
        moved to the RHS and we know that the upper bound is None after this).
        """
        body = constraint.body
        std_repn = generate_standard_repn(body)
        # make sure that we store the lower bound's value so that we need not
        # worry again during the transformation
        cons_dict = {'lower': value(constraint.lower), 'body': std_repn}
        upper = value(constraint.upper)
        constraints_to_add = [cons_dict]
        if upper is not None:
            # if it has both bounds
            if cons_dict['lower'] is not None:
                # copy the constraint and flip
                leq_side = {
                    'lower': -upper,
                    'body': generate_standard_repn(-1.0 * body),
                }
                self._move_constant_and_add_map(leq_side)
                constraints_to_add.append(leq_side)

            # If it has only an upper bound, we just need to flip it
            else:
                # just flip the constraint
                cons_dict['lower'] = -upper
                cons_dict['body'] = generate_standard_repn(-1.0 * body)
        self._move_constant_and_add_map(cons_dict)

        return constraints_to_add

    def _move_constant_and_add_map(self, cons_dict):
        """Takes constraint in dictionary form already in >= form,
        and moves the constant to the RHS
        """
        body = cons_dict['body']
        constant = value(body.constant)
        cons_dict['lower'] -= constant
        body.constant = 0

        # store a map of vars to coefficients. We can't use this in place of
        # standard repn because determinism, but this will save a lot of linear
        # time searches later. Note also that we will take the value of the
        # coefficient here so that we never have to worry about it again during
        # the transformation.
        cons_dict['map'] = ComponentMap(
            zip(body.linear_vars, [value(coef) for coef in body.linear_coefs])
        )

    def _fourier_motzkin_elimination(self, constraints, vars_to_eliminate):
        """Performs FME on the constraint list in the argument
        (which is assumed to be all >= constraints and stored in the
        dictionary representation), projecting out each of the variables in
        vars_to_eliminate"""

        # We only need to eliminate variables that actually appear in
        # this set of constraints... Revise our list.
        vars_that_appear = []
        vars_that_appear_set = ComponentSet()
        for cons in constraints:
            std_repn = cons['body']
            if not std_repn.is_linear():
                # as long as none of vars_that_appear are in the nonlinear part,
                # we are actually okay.
                nonlinear_vars = ComponentSet(
                    v for two_tuple in std_repn.quadratic_vars for v in two_tuple
                )
                nonlinear_vars.update(v for v in std_repn.nonlinear_vars)
                for var in nonlinear_vars:
                    if var in vars_to_eliminate:
                        raise RuntimeError(
                            "Variable %s appears in a nonlinear "
                            "constraint. The Fourier-Motzkin "
                            "Elimination transformation can only "
                            "be used to eliminate variables "
                            "which only appear linearly." % var.name
                        )
            for var in std_repn.linear_vars:
                if var in vars_to_eliminate:
                    if not var in vars_that_appear_set:
                        vars_that_appear.append(var)
                        vars_that_appear_set.add(var)

        # we actually begin the recursion here
        total = len(vars_that_appear)
        iteration = 1
        while vars_that_appear:
            # first var we will project out
            the_var = vars_that_appear.pop()
            logger.warning("Projecting out var %s of %s" % (iteration, total))
            if self.verbose:
                logger.info("Projecting out %s" % the_var.getname(fully_qualified=True))
                logger.info("New constraints are:")

            # we are 'reorganizing' the constraints, we sort based on the sign
            # of the coefficient of the_var: This tells us whether we have
            # the_var <= other stuff or vice versa.
            leq_list = []
            geq_list = []
            waiting_list = []

            coefs = []
            for cons in constraints:
                leaving_var_coef = cons['map'].get(the_var)
                if leaving_var_coef is None or leaving_var_coef == 0:
                    waiting_list.append(cons)
                    if self.verbose:
                        logger.info(
                            "\t%s <= %s" % (cons['lower'], cons['body'].to_expression())
                        )
                    continue

                # we know the constraint is a >= constraint, using that
                # assumption below.
                # NOTE: neither of the scalar multiplications below flip the
                # constraint. So we are sure to have only geq constraints
                # forever, which is exactly what we want.
                if not self.do_integer_arithmetic:
                    if leaving_var_coef < 0:
                        leq_list.append(
                            self._nonneg_scalar_multiply_linear_constraint(
                                cons, -1.0 / leaving_var_coef
                            )
                        )
                    else:
                        geq_list.append(
                            self._nonneg_scalar_multiply_linear_constraint(
                                cons, 1.0 / leaving_var_coef
                            )
                        )
                else:
                    coefs.append(
                        self._as_integer(
                            leaving_var_coef,
                            self._get_noninteger_coef_error_message,
                            (the_var.name, leaving_var_coef),
                        )
                    )
            if self.do_integer_arithmetic and len(coefs) > 0:
                least_common_mult = lcm(coefs)
                for cons in constraints:
                    leaving_var_coef = cons['map'].get(the_var)
                    if leaving_var_coef is None or leaving_var_coef == 0:
                        continue
                    to_lcm = least_common_mult // abs(int(leaving_var_coef))
                    if leaving_var_coef < 0:
                        leq_list.append(
                            self._nonneg_scalar_multiply_linear_constraint(cons, to_lcm)
                        )
                    else:
                        geq_list.append(
                            self._nonneg_scalar_multiply_linear_constraint(cons, to_lcm)
                        )

            constraints = waiting_list
            for leq in leq_list:
                for geq in geq_list:
                    constraints.append(self._add_linear_constraints(leq, geq))
                    if self.verbose:
                        cons = constraints[len(constraints) - 1]
                        logger.info(
                            "\t%s <= %s" % (cons['lower'], cons['body'].to_expression())
                        )

            iteration += 1

        return constraints

    def _get_noninteger_coef_error_message(self, varname, coef):
        return (
            "The do_integer_arithmetic flag was "
            "set to True, but the coefficient of "
            "%s is non-integer within the specified "
            "tolerance, with value %s. \n"
            "Please set do_integer_arithmetic="
            "False, increase integer_tolerance, "
            "or make your data integer." % (varname, coef)
        )

    def _as_integer(self, x, error_message, error_args):
        if abs(int(x) - x) <= self.integer_tolerance:
            return int(round(x))
        raise ValueError(
            error_message if error_args is None else error_message(*error_args)
        )

    def _multiply(self, scalar, coef, error_message, error_args):
        if self.do_integer_arithmetic:
            assert type(scalar) is int
            return scalar * self._as_integer(coef, error_message, error_args)
        elif abs(scalar * coef) > self.zero_tolerance:
            return scalar * coef
        else:
            return 0

    def _add(self, a, b, error_message, error_args):
        if self.do_integer_arithmetic:
            return self._as_integer(a, error_message, error_args) + self._as_integer(
                b, error_message, error_args
            )
        elif abs(a + b) > self.zero_tolerance:
            return a + b
        else:
            return 0

    def _nonneg_scalar_multiply_linear_constraint_error_msg(self, cons, coef):
        return (
            "The do_integer_arithmetic flag was set to True, but the "
            "lower bound of %s is non-integer within the specified "
            "tolerance, with value %s. \n"
            "Please set do_integer_arithmetic=False, increase "
            "integer_tolerance, or make your data integer."
            % (cons['body'].to_expression() >= cons['lower'], coef)
        )

    def _nonneg_scalar_multiply_linear_constraint(self, cons, scalar):
        """Multiplies all coefficients and the RHS of a >= constraint by scalar.
        There is no logic for flipping the equality, so this is just the
        special case with a nonnegative scalar, which is all we need.

        If self.do_integer_arithmetic is True, this assumes that scalar is an
        int. It also will throw an error if any data is non-integer (within
        tolerance)
        """
        body = cons['body']
        new_coefs = []
        for i, coef in enumerate(body.linear_coefs):
            v = body.linear_vars[i]
            new_coefs.append(
                self._multiply(
                    scalar,
                    coef,
                    self._get_noninteger_coef_error_message,
                    (v.name, coef),
                )
            )
            # update the map
            cons['map'][v] = new_coefs[i]
        body.linear_coefs = new_coefs

        body.quadratic_coefs = [scalar * coef for coef in body.quadratic_coefs]
        body.nonlinear_expr = (
            scalar * body.nonlinear_expr if body.nonlinear_expr is not None else None
        )

        # assume scalar >= 0 and constraint only has lower bound
        lb = cons['lower']
        if lb is not None:
            cons['lower'] = self._multiply(
                scalar,
                lb,
                self._nonneg_scalar_multiply_linear_constraint_error_msg,
                (cons, coef),
            )
        return cons

    def _add_linear_constraints_error_msg(self, cons1, cons2):
        return (
            "The do_integer_arithmetic flag was set to True, but while "
            "adding %s and %s, encountered a coefficient that is "
            "non-integer within the specified tolerance\n"
            "Please set do_integer_arithmetic=False, increase "
            "integer_tolerance, or make your data integer."
            % (
                cons1['body'].to_expression() >= cons1['lower'],
                cons2['body'].to_expression() >= cons2['lower'],
            )
        )

    def _add_linear_constraints(self, cons1, cons2):
        """Adds two >= constraints

        Because this is always called after
        _nonneg_scalar_multiply_linear_constraint, though it is implemented
        more generally.
        """
        ans = {'lower': None, 'body': None, 'map': ComponentMap()}
        cons1_body = cons1['body']
        cons2_body = cons2['body']

        # Need this to be both deterministic and to account for the fact that
        # Vars aren't hashable.
        all_vars = list(cons1_body.linear_vars)
        seen = ComponentSet(all_vars)
        for v in cons2_body.linear_vars:
            if v not in seen:
                all_vars.append(v)

        expr = 0
        for var in all_vars:
            coef = self._add(
                cons1['map'].get(var, 0),
                cons2['map'].get(var, 0),
                self._add_linear_constraints_error_msg,
                (cons1, cons2),
            )
            ans['map'][var] = coef
            expr += coef * var

        # deal with nonlinear stuff if there is any
        for cons in [cons1_body, cons2_body]:
            if cons.nonlinear_expr is not None:
                expr += cons.nonlinear_expr
            expr += sum(
                coef * v1 * v2
                for (coef, (v1, v2)) in zip(cons.quadratic_coefs, cons.quadratic_vars)
            )

        ans['body'] = generate_standard_repn(expr)

        # upper is None and lower exists, so this gets the constant
        ans['lower'] = self._add(
            cons1['lower'],
            cons2['lower'],
            self._add_linear_constraints_error_msg,
            (cons1, cons2),
        )

        return ans

    def post_process_fme_constraints(
        self, m, solver_factory, projected_constraints=None, tolerance=0
    ):
        """Function that solves a sequence of LPs problems to check if
        constraints are implied by each other. Deletes any that are.

        Parameters
        ----------------
        m: A model, already transformed with FME. Note that if constraints
           have been added, activated, or deactivated, we will check for
           redundancy against the whole active part of the model. If you call
           this straight after FME, you are only checking within the projected
           constraints, but otherwise it is up to the user.
        solver_factory: A SolverFactory object (constructed with a solver
                        which can solve the continuous relaxation of the
                        active constraints on the model. That is, if you
                        had nonlinear constraints unrelated to the variables
                        being projected, you need to either deactivate them or
                        provide a solver which will do the right thing.)
        projected_constraints: The ConstraintList of projected constraints.
                               Default is None, in which case we assume that
                               the FME transformation was called without
                               specifying their name, so will look for them on
                               the private transformation block.
        tolerance: Tolerance at which we decide a constraint is implied by the
                   others. Default is 0, meaning we remove the constraint if
                   the LP solve finds the constraint can be tight but not
                   violated. Setting this to a small positive value would
                   remove constraints more conservatively. Setting it to a
                   negative value would result in a relaxed problem.
        """
        if projected_constraints is None:
            # make sure m looks like what we expect
            if not hasattr(m, "_pyomo_contrib_fme_transformation"):
                raise RuntimeError(
                    "It looks like model %s has not been "
                    "transformed with the "
                    "fourier_motzkin_elimination transformation!" % m.name
                )
            transBlock = m._pyomo_contrib_fme_transformation
            if not hasattr(transBlock, 'projected_constraints'):
                raise RuntimeError(
                    "It looks the projected constraints "
                    "were manually named when the FME "
                    "transformation was called on %s. "
                    "If this is so, specify the ConstraintList "
                    "of projected constraints with the "
                    "'projected_constraints' argument." % m.name
                )
            projected_constraints = transBlock.projected_constraints

        # relax integrality so that we can do this with LP solves.
        TransformationFactory('core.relax_integer_vars').apply_to(
            m, transform_deactivated_blocks=True
        )
        # deactivate any active objectives on the model, and save what we did so
        # we can undo it after.
        active_objs = []
        for obj in m.component_data_objects(Objective, descend_into=True):
            if obj.active:
                active_objs.append(obj)
            obj.deactivate()
        # add placeholder for our own objective
        obj_name = unique_component_name(m, '_fme_post_process_obj')
        obj = Objective(expr=0)
        m.add_component(obj_name, obj)
        for i in projected_constraints:
            # If someone wants us to ignore it and leave it in the model, we
            # can.
            if not projected_constraints[i].active:
                continue
            # deactivate the constraint
            projected_constraints[i].deactivate()
            # Our constraint looks like: 0 <= a^Tx - b, so make objective to
            # maximize its infeasibility
            obj.expr = projected_constraints[i].body - projected_constraints[i].lower
            results = solver_factory.solve(m)
            if results.solver.termination_condition == TerminationCondition.unbounded:
                obj_val = -float('inf')
            elif results.solver.termination_condition != TerminationCondition.optimal:
                raise RuntimeError(
                    "Unsuccessful subproblem solve when checking"
                    "constraint %s.\n\t"
                    "Termination Condition: %s"
                    % (
                        projected_constraints[i].name,
                        results.solver.termination_condition,
                    )
                )
            else:
                obj_val = value(obj)
            # if we couldn't make it infeasible, it's useless
            if obj_val >= tolerance:
                del projected_constraints[i]
            else:
                projected_constraints[i].activate()

        # clean up
        m.del_component(obj)
        for obj in active_objs:
            obj.activate()
        # undo relax integrality
        TransformationFactory('core.relax_integer_vars').apply_to(m, undo=True)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.enums import ExtendedEnumType, IntEnum, ObjectiveSense
from pyomo.opt.results.container import MapContainer


class ProblemSense(IntEnum, metaclass=ExtendedEnumType):
    __base_enum__ = ObjectiveSense

    unknown = 0

    def __str__(self):
        return self.name


class ProblemInformation(MapContainer):
    def __init__(self):
        super().__init__()
        self.declare('name')
        self.declare('lower_bound', value=float('-inf'))
        self.declare('upper_bound', value=float('inf'))
        self.declare('number_of_objectives', value=1, required=True)
        self.declare('number_of_constraints', value=0)
        self.declare('number_of_variables', value=0)
        self.declare('number_of_binary_variables')
        self.declare('number_of_integer_variables')
        self.declare('number_of_continuous_variables')
        self.declare('number_of_nonzeros')
        self.declare('sense', value=ProblemSense.unknown, required=True)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging

from pyomo.core.base.constraint import Constraint
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.util.subsystems import TemporarySubsystemManager, generate_subsystem_blocks
from pyomo.contrib.incidence_analysis.interface import (
    IncidenceGraphInterface,
    _generate_variables_in_constraints,
)
from pyomo.contrib.incidence_analysis.config import IncidenceMethod

_log = logging.getLogger(__name__)


def generate_strongly_connected_components(
    constraints, variables=None, include_fixed=False, igraph=None
):
    """Yield in order ``BlockData`` that each contain the variables and
    constraints of a single diagonal block in a block lower triangularization
    of the incidence matrix of constraints and variables

    These diagonal blocks correspond to strongly connected components of the
    bipartite incidence graph, projected with respect to a perfect matching
    into a directed graph.

    Parameters
    ----------
    constraints: List of Pyomo constraint data objects
        Constraints used to generate strongly connected components.
    variables: List of Pyomo variable data objects
        Variables that may participate in strongly connected components.
        If not provided, all variables in the constraints will be used.
    include_fixed: Bool, optional
        Indicates whether fixed variables will be included when
        identifying variables in constraints.
    igraph: IncidenceGraphInterface, optional
        Incidence graph containing (at least) the provided constraints
        and variables.

    Yields
    ------
    Tuple of ``BlockData``, list-of-variables
        Blocks containing the variables and constraints of every strongly
        connected component, in a topological order. The variables are the
        "input variables" for that block.

    """
    if variables is None:
        variables = list(
            _generate_variables_in_constraints(
                constraints,
                include_fixed=include_fixed,
                method=IncidenceMethod.ampl_repn,
            )
        )

    if len(variables) != len(constraints):
        nvar = len(variables)
        ncon = len(constraints)
        raise RuntimeError(
            "generate_strongly_connected_components only supports systems with the"
            f" same numbers of variables and equality constraints. Got {nvar}"
            f" variables and {ncon} constraints."
        )
    if igraph is None:
        igraph = IncidenceGraphInterface()

    var_blocks, con_blocks = igraph.block_triangularize(
        variables=variables, constraints=constraints
    )
    subsets = [(cblock, vblock) for vblock, cblock in zip(var_blocks, con_blocks)]
    for block, inputs in generate_subsystem_blocks(
        subsets, include_fixed=include_fixed
    ):
        # TODO: How does len scale for reference-to-list?
        # If this assert fails, it may be due to a bug in block_triangularize
        # or generate_subsystem_block.
        assert len(block.vars) == len(block.cons)
        yield (block, inputs)


def solve_strongly_connected_components(
    block, *, solver=None, solve_kwds=None, use_calc_var=True, calc_var_kwds=None
):
    """Solve a square system of variables and equality constraints by
    solving strongly connected components individually.

    Strongly connected components (of the directed graph of constraints
    obtained from a perfect matching of variables and constraints) are
    the diagonal blocks in a block triangularization of the incidence
    matrix, so solving the strongly connected components in topological
    order is sufficient to solve the entire block.

    One-by-one blocks are solved using Pyomo's
    calculate_variable_from_constraint function, while higher-dimension
    blocks are solved using the user-provided solver object.

    Parameters
    ----------
    block: Pyomo Block
        The Pyomo block whose variables and constraints will be solved
    solver: Pyomo solver object
        The solver object that will be used to solve strongly connected
        components of size greater than one constraint. Must implement
        a solve method.
    solve_kwds: Dictionary
        Keyword arguments for the solver's solve method
    use_calc_var: Bool
        Whether to use ``calculate_variable_from_constraint`` for one-by-one
        square system solves
    calc_var_kwds: Dictionary
        Keyword arguments for calculate_variable_from_constraint

    Returns
    -------
    List of results objects returned by each call to solve

    """
    if solve_kwds is None:
        solve_kwds = {}
    if calc_var_kwds is None:
        calc_var_kwds = {}

    igraph = IncidenceGraphInterface(
        block,
        active=True,
        include_fixed=False,
        include_inequality=False,
        method=IncidenceMethod.ampl_repn,
    )
    constraints = igraph.constraints
    variables = igraph.variables

    res_list = []
    log_blocks = _log.isEnabledFor(logging.DEBUG)
    for scc, inputs in generate_strongly_connected_components(
        constraints, variables, igraph=igraph
    ):
        with TemporarySubsystemManager(to_fix=inputs, remove_bounds_on_fix=True):
            N = len(scc.vars)
            if N == 1 and use_calc_var:
                if log_blocks:
                    _log.debug(f"Solving 1x1 block: {scc.cons[0].name}.")
                results = calculate_variable_from_constraint(
                    scc.vars[0], scc.cons[0], **calc_var_kwds
                )
            else:
                if solver is None:
                    var_names = [var.name for var in scc.vars.values()][:10]
                    con_names = [con.name for con in scc.cons.values()][:10]
                    raise RuntimeError(
                        "An external solver is required if block has strongly\n"
                        "connected components of size greater than one (is not"
                        " a DAG).\nGot an SCC of size %sx%s including"
                        " components:\n%s\n%s" % (N, N, var_names, con_names)
                    )
                if log_blocks:
                    _log.debug(f"Solving {N}x{N} block.")
                results = solver.solve(scc, **solve_kwds)
            res_list.append(results)
    return res_list



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

from pyomo.contrib.cp import (
    IntervalVar,
    SequenceVar,
    no_overlap,
    first_in_sequence,
    last_in_sequence,
    alternative,
    synchronize,
)
from pyomo.contrib.cp.scheduling_expr.step_function_expressions import Step, Pulse
from pyomo.contrib.cp.repn.docplex_writer import docplex_available, LogicalToDoCplex

from pyomo.core.base.range import NumericRange
from pyomo.core.expr.numeric_expr import MinExpression, MaxExpression
from pyomo.core.expr.logical_expr import (
    equivalent,
    exactly,
    atleast,
    atmost,
    all_different,
    count_if,
)
from pyomo.core.expr.relational_expr import NotEqualExpression

from pyomo.environ import (
    ConcreteModel,
    RangeSet,
    Var,
    BooleanVar,
    Constraint,
    LogicalConstraint,
    PositiveIntegers,
    Binary,
    NonNegativeIntegers,
    NegativeIntegers,
    NonPositiveIntegers,
    Integers,
    inequality,
    Expression,
    Param,
)

try:
    import docplex.cp.model as cp

    docplex_available = True
except:
    docplex_available = False


class CommonTest(unittest.TestCase):
    def get_visitor(self):
        docplex_model = cp.CpoModel()
        return LogicalToDoCplex(docplex_model, symbolic_solver_labels=True)

    def get_model(self):
        m = ConcreteModel()
        m.I = RangeSet(10)
        m.a = Var(m.I)
        m.x = Var(within=PositiveIntegers, bounds=(6, 8))
        m.i = IntervalVar(optional=True)
        m.i2 = IntervalVar([1, 2], optional=False)

        m.b = BooleanVar()
        m.b2 = BooleanVar(['a', 'b', 'c'])

        return m


@unittest.skipIf(not docplex_available, "docplex is not available")
class TestCPExpressionWalker_AlgebraicExpressions(CommonTest):
    def test_write_addition(self):
        m = self.get_model()
        m.c = Constraint(expr=m.x + m.i.start_time + m.i2[2].length <= 3)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.x), visitor.var_map)
        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i2[2].length), visitor.var_map)

        cpx_x = visitor.var_map[id(m.x)]
        cpx_i = visitor.var_map[id(m.i)]
        cpx_i2 = visitor.var_map[id(m.i2[2])]
        self.assertTrue(
            expr[1].equals(cpx_x + cp.start_of(cpx_i) + cp.length_of(cpx_i2))
        )

        self.assertIs(visitor.pyomo_to_docplex[m.x], cpx_x)
        self.assertIs(visitor.pyomo_to_docplex[m.i], cpx_i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], cpx_i2)

    def test_write_subtraction(self):
        m = self.get_model()
        m.a.domain = Binary
        m.c = Constraint(expr=m.x - m.a[1] <= 3)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.x), visitor.var_map)
        self.assertIn(id(m.a[1]), visitor.var_map)

        x = visitor.var_map[id(m.x)]
        a1 = visitor.var_map[id(m.a[1])]

        self.assertTrue(expr[1].equals(x + (-1 * a1)))

        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertIs(visitor.pyomo_to_docplex[m.a[1]], a1)

    def test_write_product(self):
        m = self.get_model()
        m.a.domain = PositiveIntegers
        m.c = Constraint(expr=m.x * (m.a[1] + 1) <= 3)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.x), visitor.var_map)
        self.assertIn(id(m.a[1]), visitor.var_map)

        x = visitor.var_map[id(m.x)]
        a1 = visitor.var_map[id(m.a[1])]

        self.assertTrue(expr[1].equals(x * (a1 + 1)))

        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertIs(visitor.pyomo_to_docplex[m.a[1]], a1)

    def test_write_floating_point_division(self):
        m = self.get_model()
        m.a.domain = NonNegativeIntegers
        m.c = Constraint(expr=m.x / (m.a[1] + 1) <= 3)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.x), visitor.var_map)
        self.assertIn(id(m.a[1]), visitor.var_map)

        x = visitor.var_map[id(m.x)]
        a1 = visitor.var_map[id(m.a[1])]

        self.assertTrue(expr[1].equals(x / (a1 + 1)))

        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertIs(visitor.pyomo_to_docplex[m.a[1]], a1)

    def test_write_power_expression(self):
        m = self.get_model()
        m.c = Constraint(expr=m.x**2 <= 3)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.x), visitor.var_map)
        cpx_x = visitor.var_map[id(m.x)]
        # .equals checks the equality of two expressions in docplex.
        self.assertTrue(expr[1].equals(cpx_x**2))

        self.assertIs(visitor.pyomo_to_docplex[m.x], cpx_x)

    def test_write_absolute_value_expression(self):
        m = self.get_model()
        m.a.domain = NegativeIntegers
        m.c = Constraint(expr=abs(m.a[1]) + 1 <= 3)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.a[1]), visitor.var_map)

        a1 = visitor.var_map[id(m.a[1])]

        self.assertTrue(expr[1].equals(cp.abs(a1) + 1))

        self.assertIs(visitor.pyomo_to_docplex[m.a[1]], a1)

    def test_write_min_expression(self):
        m = self.get_model()
        m.a.domain = NonPositiveIntegers
        m.c = Constraint(expr=MinExpression([m.a[i] for i in m.I]) >= 3)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        a = {}
        for i in m.I:
            self.assertIn(id(m.a[i]), visitor.var_map)
            a[i] = visitor.var_map[id(m.a[i])]
            self.assertIs(visitor.pyomo_to_docplex[m.a[i]], a[i])

        self.assertTrue(expr[1].equals(cp.min(a[i] for i in m.I)))

    def test_write_max_expression(self):
        m = self.get_model()
        m.a.domain = NonPositiveIntegers
        m.c = Constraint(expr=MaxExpression([m.a[i] for i in m.I]) >= 3)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        a = {}
        for i in m.I:
            self.assertIn(id(m.a[i]), visitor.var_map)
            a[i] = visitor.var_map[id(m.a[i])]
            self.assertIs(visitor.pyomo_to_docplex[m.a[i]], a[i])

        self.assertTrue(expr[1].equals(cp.max(a[i] for i in m.I)))

    def test_expression_with_mutable_param(self):
        m = ConcreteModel()
        m.x = Var(domain=Integers, bounds=(2, 3))
        m.p = Param(initialize=4, mutable=True)
        e = m.p * m.x

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]

        self.assertTrue(expr[1].equals(4 * x))

    def test_monomial_expressions(self):
        m = ConcreteModel()
        m.x = Var(domain=Integers, bounds=(1, 4))
        m.p = Param(initialize=4, mutable=True)

        visitor = self.get_visitor()

        const_expr = 3 * m.x
        nested_expr = (1 / m.p) * m.x
        pow_expr = (m.p ** (0.5)) * m.x

        e = m.x * 4
        expr = visitor.walk_expression((e, e, 0))
        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        self.assertTrue(expr[1].equals(4 * x))

        e = 1.0 * m.x
        expr = visitor.walk_expression((e, e, 0))
        self.assertTrue(expr[1].equals(x))

        e = (1 / m.p) * m.x
        expr = visitor.walk_expression((e, e, 0))
        self.assertTrue(expr[1].equals(cp.float_div(1, 4) * x))

        e = (m.p ** (0.5)) * m.x
        expr = visitor.walk_expression((e, e, 0))
        self.assertTrue(expr[1].equals(cp.power(4, 0.5) * x))


@unittest.skipIf(not docplex_available, "docplex is not available")
class TestCPExpressionWalker_LogicalExpressions(CommonTest):
    def test_write_logical_and(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.b.land(m.b2['b']))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.b2['b']), visitor.var_map)

        b = visitor.var_map[id(m.b)]
        b2b = visitor.var_map[id(m.b2['b'])]

        self.assertTrue(expr[1].equals(cp.logical_and(b, b2b)))

        # ESJ: This is ludicrous, but I don't know how to get the args of a CP
        # expression, so testing that we were correct in the pyomo to docplex
        # map by checking that we can build an expression that is the same as b
        # (because b is actually "b == 1" since docplex doesn't believe in
        # Booleans)
        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))
        self.assertTrue(b2b.equals(visitor.pyomo_to_docplex[m.b2['b']] == 1))

    def test_write_logical_or(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.b.lor(m.i.is_present))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.i), visitor.var_map)
        b = visitor.var_map[id(m.b)]
        i = visitor.var_map[id(m.i)]

        self.assertTrue(expr[1].equals(cp.logical_or(b, cp.presence_of(i))))

        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

    def test_write_xor(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.b.xor(m.i2[2].start_time >= 5))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        b = visitor.var_map[id(m.b)]
        i22 = visitor.var_map[id(m.i2[2])]

        # [ESJ 9/22/22]: This isn't the greatest test because there's no direct
        # translation so how we choose to represent this could change.
        self.assertTrue(
            expr[1].equals(cp.count([b, cp.less_or_equal(5, cp.start_of(i22))], 1) == 1)
        )

        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)

    def test_write_logical_not(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=~m.b2['a'])
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.b2['a']), visitor.var_map)
        b2a = visitor.var_map[id(m.b2['a'])]

        self.assertTrue(expr[1].equals(cp.logical_not(b2a)))

        self.assertTrue(b2a.equals(visitor.pyomo_to_docplex[m.b2['a']] == 1))

    def test_equivalence(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=equivalent(~m.b2['a'], m.b))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.b2['a']), visitor.var_map)
        b = visitor.var_map[id(m.b)]
        b2a = visitor.var_map[id(m.b2['a'])]

        self.assertTrue(expr[1].equals(cp.equal(cp.logical_not(b2a), b)))

        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))
        self.assertTrue(b2a.equals(visitor.pyomo_to_docplex[m.b2['a']] == 1))

    def test_equality(self):
        m = self.get_model()
        m.a.domain = Integers
        m.c = LogicalConstraint(expr=m.b.implies(m.a[3] == 4))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.a[3]), visitor.var_map)
        b = visitor.var_map[id(m.b)]
        a3 = visitor.var_map[id(m.a[3])]

        self.assertTrue(expr[1].equals(cp.if_then(b, cp.equal(a3, 4))))

        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))
        self.assertIs(visitor.pyomo_to_docplex[m.a[3]], a3)

    def test_inequality(self):
        m = self.get_model()
        m.a.domain = Integers
        m.c = LogicalConstraint(expr=m.b.implies(m.a[3] >= m.a[4]))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.a[3]), visitor.var_map)
        self.assertIn(id(m.a[4]), visitor.var_map)
        b = visitor.var_map[id(m.b)]
        a3 = visitor.var_map[id(m.a[3])]
        a4 = visitor.var_map[id(m.a[4])]

        self.assertTrue(expr[1].equals(cp.if_then(b, cp.less_or_equal(a4, a3))))

        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))
        self.assertIs(visitor.pyomo_to_docplex[m.a[3]], a3)
        self.assertIs(visitor.pyomo_to_docplex[m.a[4]], a4)

    def test_ranged_inequality(self):
        m = self.get_model()
        m.a.domain = Integers
        m.c = Constraint(expr=inequality(3, m.a[2], 5))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.a[2]), visitor.var_map)
        a2 = visitor.var_map[id(m.a[2])]

        self.assertTrue(expr[1].equals(cp.range(a2, 3, 5)))

    def test_not_equal(self):
        m = self.get_model()
        m.a.domain = Integers
        m.c = LogicalConstraint(expr=m.b.implies(NotEqualExpression([m.a[3], m.a[4]])))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.a[3]), visitor.var_map)
        self.assertIn(id(m.a[4]), visitor.var_map)
        b = visitor.var_map[id(m.b)]
        a3 = visitor.var_map[id(m.a[3])]
        a4 = visitor.var_map[id(m.a[4])]

        self.assertTrue(expr[1].equals(cp.if_then(b, a3 != a4)))

        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))
        self.assertIs(visitor.pyomo_to_docplex[m.a[3]], a3)
        self.assertIs(visitor.pyomo_to_docplex[m.a[4]], a4)

    def test_exactly_expression(self):
        m = self.get_model()
        m.a.domain = Integers
        m.c = LogicalConstraint(expr=exactly(3, [m.a[i] == 4 for i in m.I]))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        a = {}
        for i in m.I:
            self.assertIn(id(m.a[i]), visitor.var_map)
            a[i] = visitor.var_map[id(m.a[i])]
            self.assertIs(visitor.pyomo_to_docplex[m.a[i]], a[i])

        self.assertTrue(
            expr[1].equals(cp.equal(cp.count([a[i] == 4 for i in m.I], 1), 3))
        )

    def test_atleast_expression(self):
        m = self.get_model()
        m.a.domain = Integers
        m.c = LogicalConstraint(expr=atleast(3, [m.a[i] == 4 for i in m.I]))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        a = {}
        for i in m.I:
            self.assertIn(id(m.a[i]), visitor.var_map)
            a[i] = visitor.var_map[id(m.a[i])]
            self.assertIs(visitor.pyomo_to_docplex[m.a[i]], a[i])

        self.assertTrue(
            expr[1].equals(
                cp.greater_or_equal(cp.count([a[i] == 4 for i in m.I], 1), 3)
            )
        )

    def test_atmost_expression(self):
        m = self.get_model()
        m.a.domain = Integers
        m.c = LogicalConstraint(expr=atmost(3, [m.a[i] == 4 for i in m.I]))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        a = {}
        for i in m.I:
            self.assertIn(id(m.a[i]), visitor.var_map)
            a[i] = visitor.var_map[id(m.a[i])]
            self.assertIs(visitor.pyomo_to_docplex[m.a[i]], a[i])

        self.assertTrue(
            expr[1].equals(cp.less_or_equal(cp.count([a[i] == 4 for i in m.I], 1), 3))
        )

    def test_all_diff_expression(self):
        m = self.get_model()
        m.a.domain = Integers
        m.a.bounds = (11, 20)
        m.c = LogicalConstraint(expr=all_different(m.a))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        a = {}
        for i in m.I:
            self.assertIn(id(m.a[i]), visitor.var_map)
            a[i] = visitor.var_map[id(m.a[i])]
            self.assertIs(visitor.pyomo_to_docplex[m.a[i]], a[i])

        self.assertTrue(expr[1].equals(cp.all_diff(a[i] for i in m.I)))

    def test_count_if_expression(self):
        m = self.get_model()
        m.a.domain = Integers
        m.a.bounds = (11, 20)
        m.c = Constraint(expr=count_if(m.a[i] == i for i in m.I) == 5)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        a = {}
        for i in m.I:
            self.assertIn(id(m.a[i]), visitor.var_map)
            a[i] = visitor.var_map[id(m.a[i])]
            self.assertIs(visitor.pyomo_to_docplex[m.a[i]], a[i])

        self.assertTrue(expr[1].equals(cp.count((a[i] == i for i in m.I), 1) == 5))

    def test_interval_var_is_present(self):
        m = self.get_model()
        m.a.domain = Integers
        m.c = LogicalConstraint(expr=m.i.is_present.implies(m.a[1] == 5))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.a[1]), visitor.var_map)
        self.assertIn(id(m.i), visitor.var_map)
        a1 = visitor.var_map[id(m.a[1])]
        i = visitor.var_map[id(m.i)]

        self.assertTrue(expr[1].equals(cp.if_then(cp.presence_of(i), a1 == 5)))

        self.assertIs(visitor.pyomo_to_docplex[m.a[1]], a1)
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

    def test_interval_var_is_present_indirection(self):
        m = self.get_model()
        m.a.domain = Integers
        m.y = Var(domain=Integers, bounds=[1, 2])

        m.c = LogicalConstraint(expr=m.i2[m.y].is_present.implies(m.a[1] >= 7))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.a[1]), visitor.var_map)
        a1 = visitor.var_map[id(m.a[1])]

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]

        self.assertTrue(
            expr[1].equals(
                cp.if_then(
                    cp.element(
                        [cp.presence_of(i21), cp.presence_of(i22)], 0 + 1 * (y - 1) // 1
                    )
                    == True,
                    cp.less_or_equal(7, a1),
                )
            )
        )

        self.assertIs(visitor.pyomo_to_docplex[m.a[1]], a1)
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)

    def test_is_present_indirection_and_length(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=[1, 2])

        m.c = LogicalConstraint(expr=m.i2[m.y].is_present.land(m.i2[m.y].length >= 7))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]

        self.assertTrue(
            expr[1].equals(
                cp.logical_and(
                    cp.element(
                        [cp.presence_of(i21), cp.presence_of(i22)], 0 + 1 * (y - 1) // 1
                    )
                    == True,
                    cp.less_or_equal(
                        7,
                        cp.element(
                            [cp.length_of(i21), cp.length_of(i22)], 0 + 1 * (y - 1) // 1
                        ),
                    ),
                )
            )
        )

        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)

    def test_handle_getattr_lor(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=(1, 2))

        e = m.i2[m.y].is_present.lor(~m.b)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.b), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        b = visitor.var_map[id(m.b)]

        self.assertTrue(
            expr[1].equals(
                cp.logical_or(
                    cp.element(
                        [cp.presence_of(i21), cp.presence_of(i22)], 0 + 1 * (y - 1) // 1
                    )
                    == True,
                    cp.logical_not(b),
                )
            )
        )

        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))

    def test_handle_getattr_xor(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=(1, 2))

        e = m.i2[m.y].is_present.xor(m.b)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.b), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        b = visitor.var_map[id(m.b)]

        self.assertTrue(
            expr[1].equals(
                cp.equal(
                    cp.count(
                        [
                            cp.element(
                                [cp.presence_of(i21), cp.presence_of(i22)],
                                0 + 1 * (y - 1) // 1,
                            )
                            == True,
                            b,
                        ],
                        1,
                    ),
                    1,
                )
            )
        )

        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))

    def test_handle_getattr_equivalent_to(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=(1, 2))

        e = m.i2[m.y].is_present.equivalent_to(~m.b)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.b), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        b = visitor.var_map[id(m.b)]

        self.assertTrue(
            expr[1].equals(
                cp.equal(
                    cp.element(
                        [cp.presence_of(i21), cp.presence_of(i22)], 0 + 1 * (y - 1) // 1
                    )
                    == True,
                    cp.logical_not(b),
                )
            )
        )

        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))

    def test_logical_or_on_indirection(self):
        m = ConcreteModel()
        m.b = BooleanVar([2, 3, 4, 5])
        m.x = Var(domain=Integers, bounds=(3, 5))

        e = m.b[m.x].lor(m.x == 5)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.x), visitor.var_map)
        self.assertIn(id(m.b[3]), visitor.var_map)
        self.assertIn(id(m.b[4]), visitor.var_map)
        self.assertIn(id(m.b[5]), visitor.var_map)

        x = visitor.var_map[id(m.x)]
        b3 = visitor.var_map[id(m.b[3])]
        b4 = visitor.var_map[id(m.b[4])]
        b5 = visitor.var_map[id(m.b[5])]

        self.assertTrue(
            expr[1].equals(
                cp.logical_or(
                    cp.element([b3, b4, b5], 0 + 1 * (x - 3) // 1) == True,
                    cp.equal(x, 5),
                )
            )
        )

        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertTrue(b3.equals(visitor.pyomo_to_docplex[m.b[3]] == 1))
        self.assertTrue(b4.equals(visitor.pyomo_to_docplex[m.b[4]] == 1))
        self.assertTrue(b5.equals(visitor.pyomo_to_docplex[m.b[5]] == 1))

    def test_logical_xor_on_indirection(self):
        m = ConcreteModel()
        m.b = BooleanVar([2, 3, 4, 5])
        m.b[4].fix(False)
        m.x = Var(domain=Integers, bounds=(3, 5))

        e = m.b[m.x].xor(m.x == 5)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.x), visitor.var_map)
        self.assertIn(id(m.b[3]), visitor.var_map)
        self.assertIn(id(m.b[5]), visitor.var_map)

        x = visitor.var_map[id(m.x)]
        b3 = visitor.var_map[id(m.b[3])]
        b5 = visitor.var_map[id(m.b[5])]

        self.assertTrue(
            expr[1].equals(
                cp.equal(
                    cp.count(
                        [
                            cp.element([b3, False, b5], 0 + 1 * (x - 3) // 1) == True,
                            cp.equal(x, 5),
                        ],
                        1,
                    ),
                    1,
                )
            )
        )

        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertTrue(b3.equals(visitor.pyomo_to_docplex[m.b[3]] == 1))
        self.assertTrue(b5.equals(visitor.pyomo_to_docplex[m.b[5]] == 1))

    def test_using_precedence_expr_as_boolean_expr(self):
        m = self.get_model()
        e = m.b.implies(m.i2[2].start_time.before(m.i2[1].start_time))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)

        b = visitor.var_map[id(m.b)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]

        self.assertTrue(
            expr[1].equals(cp.if_then(b, cp.start_of(i22) + 0 <= cp.start_of(i21)))
        )

        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))

    def test_using_precedence_expr_as_boolean_expr_positive_delay(self):
        m = self.get_model()
        e = m.b.implies(m.i2[2].start_time.before(m.i2[1].start_time, delay=4))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)

        b = visitor.var_map[id(m.b)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]

        self.assertTrue(
            expr[1].equals(cp.if_then(b, cp.start_of(i22) + 4 <= cp.start_of(i21)))
        )

        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))

    def test_using_precedence_expr_as_boolean_expr_negative_delay(self):
        m = self.get_model()
        e = m.b.implies(m.i2[2].start_time.at(m.i2[1].start_time, delay=-3))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.b), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)

        b = visitor.var_map[id(m.b)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]

        self.assertTrue(
            expr[1].equals(cp.if_then(b, cp.start_of(i22) + (-3) == cp.start_of(i21)))
        )

        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertTrue(b.equals(visitor.pyomo_to_docplex[m.b] == 1))


@unittest.skipIf(not docplex_available, "docplex is not available")
class TestCPExpressionWalker_IntervalVars(CommonTest):
    def test_interval_var_fixed_presences_correct(self):
        m = self.get_model()

        m.silly = LogicalConstraint(expr=m.i.is_present)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.silly.expr, m.silly, 0))
        self.assertIn(id(m.i), visitor.var_map)
        i = visitor.var_map[id(m.i)]
        # Check that docplex knows it's optional
        self.assertTrue(i.is_optional())
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        # Now fix it to absent
        m.i.is_present.fix(False)
        m.c = LogicalConstraint(expr=m.i.is_present.lor(m.i2[1].start_time == 2))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.i2[1]), visitor.var_map)
        i21 = visitor.var_map[id(m.i2[1])]
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIn(id(m.i), visitor.var_map)
        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        # Check that we passed on the presence info to docplex
        self.assertTrue(i.is_absent())
        self.assertTrue(i21.is_present())
        # Not testing the expression here because sometime we might optimize out
        # the presence_of call for fixed absent vars, but for now I haven't.

    def test_interval_var_fixed_length(self):
        m = ConcreteModel()
        m.i = IntervalVar(start=(2, 7), end=(6, 11), optional=True)
        m.i.length.fix(4)
        m.silly = LogicalConstraint(expr=m.i.is_present)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.silly.expr, m.silly, 0))

        self.assertIn(id(m.i), visitor.var_map)
        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        self.assertTrue(i.is_optional())
        self.assertEqual(i.get_length(), (4, 4))
        self.assertEqual(i.get_start(), (2, 7))
        self.assertEqual(i.get_end(), (6, 11))

    def test_interval_var_fixed_start_and_end(self):
        m = ConcreteModel()
        m.i = IntervalVar(start=(3, 7), end=(6, 10))
        m.i.start_time.fix(3)
        m.i.end_time.fix(6)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.i, m.i, 0))

        self.assertIn(id(m.i), visitor.var_map)
        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        self.assertFalse(i.is_optional())
        self.assertEqual(i.get_start(), (3, 3))
        self.assertEqual(i.get_end(), (6, 6))


@unittest.skipIf(not docplex_available, "docplex is not available")
class TestCPExpressionWalker_SequenceVars(CommonTest):
    def get_model(self):
        m = super().get_model()
        m.seq = SequenceVar(expr=[m.i, m.i2[1], m.i2[2]])

        return m

    def check_scalar_sequence_var(self, m, visitor):
        self.assertIn(id(m.seq), visitor.var_map)
        seq = visitor.var_map[id(m.seq)]
        self.assertIs(visitor.pyomo_to_docplex[m.seq], seq)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)

        ivs = seq.get_interval_variables()
        self.assertEqual(len(ivs), 3)
        self.assertIs(ivs[0], i)
        self.assertIs(ivs[1], i21)
        self.assertIs(ivs[2], i22)

        return seq, i, i21, i22

    def test_scalar_sequence_var(self):
        m = self.get_model()

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.seq, m.seq, 0))
        self.check_scalar_sequence_var(m, visitor)

    def test_no_overlap(self):
        m = self.get_model()
        e = no_overlap(m.seq)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        seq, i, i21, i22 = self.check_scalar_sequence_var(m, visitor)
        self.assertTrue(expr[1].equals(cp.no_overlap(seq)))

    def test_first_in_sequence(self):
        m = self.get_model()
        e = first_in_sequence(m.i2[1], m.seq)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        seq, i, i21, i22 = self.check_scalar_sequence_var(m, visitor)
        self.assertTrue(expr[1].equals(cp.first(seq, i21)))

    def test_before_in_sequence(self):
        m = self.get_model()
        e = last_in_sequence(m.i, m.seq)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        seq, i, i21, i22 = self.check_scalar_sequence_var(m, visitor)
        self.assertTrue(expr[1].equals(cp.last(seq, i)))

    def test_last_in_sequence(self):
        m = self.get_model()
        e = last_in_sequence(m.i2[1], m.seq)
        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        seq, i, i21, i22 = self.check_scalar_sequence_var(m, visitor)
        self.assertTrue(expr[1].equals(cp.last(seq, i21)))


@unittest.skipIf(not docplex_available, "docplex is not available")
class TestCPExpressionWalker_PrecedenceExpressions(CommonTest):
    def test_start_before_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.i.start_time.before(m.i2[1].start_time))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)

        self.assertTrue(expr[1].equals(cp.start_before_start(i, i21, 0)))

    def test_start_before_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.i.start_time.before(m.i2[1].end_time, delay=3))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)

        self.assertTrue(expr[1].equals(cp.start_before_end(i, i21, 3)))

    def test_end_before_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.i.end_time.before(m.i2[1].start_time, delay=-2))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)

        self.assertTrue(expr[1].equals(cp.end_before_start(i, i21, -2)))

    def test_end_before_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.i.end_time.before(m.i2[1].end_time, delay=6))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)

        self.assertTrue(expr[1].equals(cp.end_before_end(i, i21, 6)))

    def test_start_at_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.i.start_time.at(m.i2[1].start_time))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)

        self.assertTrue(expr[1].equals(cp.start_at_start(i, i21, 0)))

    def test_start_at_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.i.start_time.at(m.i2[1].end_time, delay=3))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)

        self.assertTrue(expr[1].equals(cp.start_at_end(i, i21, 3)))

    def test_end_at_start(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.i.end_time.at(m.i2[1].start_time, delay=-2))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)

        self.assertTrue(expr[1].equals(cp.end_at_start(i, i21, -2)))

    def test_end_at_end(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.i.end_time.at(m.i2[1].end_time, delay=6))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)

        self.assertTrue(expr[1].equals(cp.end_at_end(i, i21, 6)))

    ##
    # Tests for precedence constraints with indirection
    ##

    def test_indirection_before_constraint(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=[1, 2])
        m.c = LogicalConstraint(expr=m.i2[m.y].start_time.before(m.i.end_time, delay=3))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        self.assertTrue(
            expr[1].equals(
                cp.element([cp.start_of(i21), cp.start_of(i22)], 0 + 1 * (y - 1) // 1)
                + 3
                <= cp.end_of(i)
            )
        )

    def test_indirection_after_constraint(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=[1, 2])
        m.c = LogicalConstraint(expr=m.i2[m.y].start_time.after(m.i.end_time, delay=-2))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        self.assertTrue(
            expr[1].equals(
                cp.end_of(i) + (-2)
                <= cp.element(
                    [cp.start_of(i21), cp.start_of(i22)], 0 + 1 * (y - 1) // 1
                )
            )
        )

    def test_indirection_at_constraint(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=[1, 2])
        m.c = LogicalConstraint(expr=m.i2[m.y].start_time.at(m.i.end_time, delay=4))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        self.assertTrue(
            expr[1].equals(
                cp.element([cp.start_of(i21), cp.start_of(i22)], 0 + 1 * (y - 1) // 1)
                == cp.end_of(i) + 4
            )
        )

    def test_before_indirection_constraint(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=[1, 2])
        m.c = LogicalConstraint(
            expr=m.i.start_time.before(m.i2[m.y].end_time, delay=-4)
        )

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        self.assertTrue(
            expr[1].equals(
                cp.start_of(i) + (-4)
                <= cp.element([cp.end_of(i21), cp.end_of(i22)], 0 + 1 * (y - 1) // 1)
            )
        )

    def test_after_indirection_constraint(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=[1, 2])
        m.c = LogicalConstraint(expr=m.i.start_time.after(m.i2[m.y].end_time))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        self.assertTrue(
            expr[1].equals(
                cp.element([cp.end_of(i21), cp.end_of(i22)], 0 + 1 * (y - 1) // 1) + 0
                <= cp.start_of(i)
            )
        )

    def test_at_indirection_constraint(self):
        m = self.get_model()
        m.y = Var(domain=Integers, bounds=[1, 2])
        m.c = LogicalConstraint(expr=m.i.start_time.at(m.i2[m.y].end_time, delay=-6))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        self.assertTrue(
            expr[1].equals(
                cp.start_of(i) + (-6)
                == cp.element([cp.end_of(i21), cp.end_of(i22)], 0 + 1 * (y - 1) // 1)
            )
        )

    def test_double_indirection_before_constraint(self):
        m = self.get_model()
        # add interval var x can index
        m.i3 = IntervalVar([(1, 3), (1, 4), (1, 5)], length=4, optional=True)
        m.y = Var(domain=Integers, bounds=[1, 2])
        m.c = LogicalConstraint(
            expr=m.i3[1, m.x - 3].start_time.before(m.i2[m.y].end_time)
        )

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i3[1, 3]), visitor.var_map)
        self.assertIn(id(m.i3[1, 4]), visitor.var_map)
        self.assertIn(id(m.i3[1, 5]), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        x = visitor.var_map[id(m.x)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        i33 = visitor.var_map[id(m.i3[1, 3])]
        i34 = visitor.var_map[id(m.i3[1, 4])]
        i35 = visitor.var_map[id(m.i3[1, 5])]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertIs(visitor.pyomo_to_docplex[m.i3[1, 3]], i33)
        self.assertIs(visitor.pyomo_to_docplex[m.i3[1, 4]], i34)
        self.assertIs(visitor.pyomo_to_docplex[m.i3[1, 5]], i35)

        self.assertTrue(
            expr[1].equals(
                cp.element(
                    [cp.start_of(i33), cp.start_of(i34), cp.start_of(i35)],
                    0 + 1 * (x + (-3) - 3) // 1,
                )
                <= cp.element([cp.end_of(i21), cp.end_of(i22)], 0 + 1 * (y - 1) // 1)
            )
        )

    def test_double_indirection_after_constraint(self):
        m = self.get_model()
        # add interval var x can index
        m.i3 = IntervalVar([(1, 3), (1, 4), (1, 5)], length=4, optional=True)
        m.y = Var(domain=Integers, bounds=[1, 2])
        m.c = LogicalConstraint(
            expr=m.i3[1, m.x - 3].start_time.after(m.i2[m.y].end_time)
        )

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i3[1, 3]), visitor.var_map)
        self.assertIn(id(m.i3[1, 4]), visitor.var_map)
        self.assertIn(id(m.i3[1, 5]), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        x = visitor.var_map[id(m.x)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        i33 = visitor.var_map[id(m.i3[1, 3])]
        i34 = visitor.var_map[id(m.i3[1, 4])]
        i35 = visitor.var_map[id(m.i3[1, 5])]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertIs(visitor.pyomo_to_docplex[m.i3[1, 3]], i33)
        self.assertIs(visitor.pyomo_to_docplex[m.i3[1, 4]], i34)
        self.assertIs(visitor.pyomo_to_docplex[m.i3[1, 5]], i35)

        self.assertTrue(
            expr[1].equals(
                cp.element([cp.end_of(i21), cp.end_of(i22)], 0 + 1 * (y - 1) // 1)
                <= cp.element(
                    [cp.start_of(i33), cp.start_of(i34), cp.start_of(i35)],
                    0 + 1 * (x + (-3) - 3) // 1,
                )
            )
        )

    def test_double_indirection_at_constraint(self):
        m = self.get_model()
        # add interval var x can index
        m.i3 = IntervalVar([(1, 3), (1, 4), (1, 5)], length=4, optional=True)
        m.y = Var(domain=Integers, bounds=[1, 2])
        m.c = LogicalConstraint(expr=m.i3[1, m.x - 3].start_time.at(m.i2[m.y].end_time))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.y), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)
        self.assertIn(id(m.i3[1, 3]), visitor.var_map)
        self.assertIn(id(m.i3[1, 4]), visitor.var_map)
        self.assertIn(id(m.i3[1, 5]), visitor.var_map)

        y = visitor.var_map[id(m.y)]
        x = visitor.var_map[id(m.x)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        i33 = visitor.var_map[id(m.i3[1, 3])]
        i34 = visitor.var_map[id(m.i3[1, 4])]
        i35 = visitor.var_map[id(m.i3[1, 5])]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)
        self.assertIs(visitor.pyomo_to_docplex[m.i3[1, 3]], i33)
        self.assertIs(visitor.pyomo_to_docplex[m.i3[1, 4]], i34)
        self.assertIs(visitor.pyomo_to_docplex[m.i3[1, 5]], i35)

        self.assertTrue(
            expr[1].equals(
                cp.element(
                    [cp.start_of(i33), cp.start_of(i34), cp.start_of(i35)],
                    0 + 1 * (x + (-3) - 3) // 1,
                )
                == cp.element([cp.end_of(i21), cp.end_of(i22)], 0 + 1 * (y - 1) // 1)
            )
        )

    def test_indirection_nonconstant_step_size(self):
        m = ConcreteModel()

        def param_rule(m, i):
            return i + 1

        m.p = Param([1, 3, 4], initialize=param_rule)
        m.x = Var(within={1, 3, 4})
        e = m.p[m.x]

        visitor = self.get_visitor()
        with self.assertRaisesRegex(
            ValueError,
            r"Variable indirection 'p\[x\]' is over a discrete domain "
            "without a constant step size. This is not supported.",
        ):
            expr = visitor.walk_expression((e, e, 0))

    def test_indirection_with_param(self):
        m = ConcreteModel()

        def param_rule(m, i):
            return i + 1

        m.p = Param([1, 3, 5], initialize=param_rule)
        m.x = Var(within={1, 3, 5})
        m.a = Var(domain=Integers, bounds=(0, 100))

        e = m.p[m.x] / m.a
        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.x), visitor.var_map)
        self.assertIn(id(m.a), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        a = visitor.var_map[id(m.a)]
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertIs(visitor.pyomo_to_docplex[m.a], a)

        self.assertTrue(expr[1].equals(cp.element([2, 4, 6], 0 + 1 * (x - 1) // 2) / a))


@unittest.skipIf(not docplex_available, "docplex is not available")
class TestCPExpressionWalker_HierarchicalScheduling(CommonTest):
    def get_model(self):
        m = ConcreteModel()

        def start_rule(m, i):
            return 2 * i

        def length_rule(m, i):
            return i

        m.iv = IntervalVar(
            [1, 2, 3], start=start_rule, length=length_rule, optional=True
        )
        m.whole_enchilada = IntervalVar()

        return m

    def test_spans(self):
        m = self.get_model()
        e = m.whole_enchilada.spans(m.iv[i] for i in [1, 2, 3])

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.whole_enchilada), visitor.var_map)
        whole_enchilada = visitor.var_map[id(m.whole_enchilada)]
        self.assertIs(visitor.pyomo_to_docplex[m.whole_enchilada], whole_enchilada)

        iv = {}
        for i in [1, 2, 3]:
            self.assertIn(id(m.iv[i]), visitor.var_map)
            iv[i] = visitor.var_map[id(m.iv[i])]

        self.assertTrue(
            expr[1].equals(cp.span(whole_enchilada, [iv[i] for i in [1, 2, 3]]))
        )

    def test_alternative(self):
        m = self.get_model()
        e = alternative(m.whole_enchilada, [m.iv[i] for i in [1, 2, 3]])

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.whole_enchilada), visitor.var_map)
        whole_enchilada = visitor.var_map[id(m.whole_enchilada)]
        self.assertIs(visitor.pyomo_to_docplex[m.whole_enchilada], whole_enchilada)

        iv = {}
        for i in [1, 2, 3]:
            self.assertIn(id(m.iv[i]), visitor.var_map)
            iv[i] = visitor.var_map[id(m.iv[i])]

        self.assertTrue(
            expr[1].equals(cp.alternative(whole_enchilada, [iv[i] for i in [1, 2, 3]]))
        )

    def test_synchronize(self):
        m = self.get_model()
        e = synchronize(m.whole_enchilada, [m.iv[i] for i in [1, 2, 3]])

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        self.assertIn(id(m.whole_enchilada), visitor.var_map)
        whole_enchilada = visitor.var_map[id(m.whole_enchilada)]
        self.assertIs(visitor.pyomo_to_docplex[m.whole_enchilada], whole_enchilada)

        iv = {}
        for i in [1, 2, 3]:
            self.assertIn(id(m.iv[i]), visitor.var_map)
            iv[i] = visitor.var_map[id(m.iv[i])]

        self.assertTrue(
            expr[1].equals(cp.synchronize(whole_enchilada, [iv[i] for i in [1, 2, 3]]))
        )


@unittest.skipIf(not docplex_available, "docplex is not available")
class TestCPExpressionWalker_CumulFuncExpressions(CommonTest):
    def test_always_in(self):
        m = self.get_model()
        f = (
            Pulse((m.i, 3))
            + Step(m.i2[1].start_time, height=2)
            - Step(m.i2[2].end_time, height=-1)
            + Step(3, height=4)
        )
        m.c = LogicalConstraint(expr=f.within((0, 3), (0, 10)))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)
        self.assertIn(id(m.i2[1]), visitor.var_map)
        self.assertIn(id(m.i2[2]), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        i21 = visitor.var_map[id(m.i2[1])]
        i22 = visitor.var_map[id(m.i2[2])]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[1]], i21)
        self.assertIs(visitor.pyomo_to_docplex[m.i2[2]], i22)

        self.assertTrue(
            expr[1].equals(
                cp.always_in(
                    cp.pulse(i, 3)
                    + cp.step_at_start(i21, 2)
                    - cp.step_at_end(i22, -1)
                    + cp.step_at(3, 4),
                    interval=(0, 10),
                    min=0,
                    max=3,
                )
            )
        )

    def test_always_in_single_pulse(self):
        # This is a bit silly as you can tell whether or not it is feasible
        # structurally, but there's no reason it couldn't happen.
        m = self.get_model()
        f = Pulse((m.i, 3))
        m.c = LogicalConstraint(expr=f.within((0, 3), (0, 10)))
        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.i), visitor.var_map)

        i = visitor.var_map[id(m.i)]
        self.assertIs(visitor.pyomo_to_docplex[m.i], i)

        self.assertTrue(
            expr[1].equals(cp.always_in(cp.pulse(i, 3), interval=(0, 10), min=0, max=3))
        )


@unittest.skipIf(not docplex_available, "docplex is not available")
class TestCPExpressionWalker_NamedExpressions(CommonTest):
    def test_named_expression(self):
        m = self.get_model()
        m.e = Expression(expr=m.x**2 + 7)
        m.c = Constraint(expr=m.e <= 32)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)

        self.assertTrue(expr[1].equals(x**2 + 7))

    def test_repeated_named_expression(self):
        m = self.get_model()
        m.e = Expression(expr=m.x**2 + 7)
        m.c = Constraint(expr=m.e - 8 * m.e <= 32)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)

        self.assertTrue(expr[1].equals(x**2 + 7 + (-1) * (8 * (x**2 + 7))))


@unittest.skipIf(not docplex_available, "docplex is not available")
class TestCPExpressionWalker_Vars(CommonTest):
    def test_complain_about_non_integer_vars(self):
        m = self.get_model()
        m.c = LogicalConstraint(expr=m.i.is_present.implies(m.a[1] == 5))

        visitor = self.get_visitor()
        with self.assertRaisesRegex(
            ValueError,
            "The LogicalToDoCplex writer can only support integer- or "
            r"Boolean-valued variables. Cannot write Var 'a\[1\]' with "
            "domain 'Reals'",
        ):
            expr = visitor.walk_expression((m.c.expr, m.c, 0))

    def test_fixed_integer_var(self):
        m = self.get_model()
        m.a.domain = Integers
        m.a[1].fix(3)
        m.c = Constraint(expr=m.a[1] + m.a[2] >= 4)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.a[2]), visitor.var_map)
        a2 = visitor.var_map[id(m.a[2])]
        self.assertIs(visitor.pyomo_to_docplex[m.a[2]], a2)

        self.assertTrue(expr[1].equals(3 + a2))

    def test_fixed_boolean_var(self):
        m = self.get_model()
        m.b.fix(False)
        m.b2['a'].fix(True)
        m.c = LogicalConstraint(expr=m.b.lor(m.b2['a'].land(m.b2['b'])))

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.expr, m.c, 0))

        self.assertIn(id(m.b2['b']), visitor.var_map)
        b2b = visitor.var_map[id(m.b2['b'])]
        self.assertTrue(b2b.equals(visitor.pyomo_to_docplex[m.b2['b']] == 1))

        self.assertTrue(expr[1].equals(cp.logical_or(False, cp.logical_and(True, b2b))))

    def test_indirection_single_index(self):
        m = self.get_model()
        m.a.domain = Integers
        m.c = Constraint(expr=m.a[m.x] >= 3.5)

        visitor = self.get_visitor()
        expr = visitor.walk_expression((m.c.body, m.c, 0))

        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        a = []
        # only need indices 6, 7, and 8 from a, since that's what x is capable
        # of selecting.
        for idx in [6, 7, 8]:
            v = m.a[idx]
            self.assertIn(id(v), visitor.var_map)
            cpx_v = visitor.var_map[id(v)]
            self.assertIs(visitor.pyomo_to_docplex[v], cpx_v)
            a.append(cpx_v)
        # since x is between 6 and 8, we subtract 6 from it for it to be the
        # right index
        self.assertTrue(expr[1].equals(cp.element(a, 0 + 1 * (x - 6) // 1)))

    def test_indirection_multi_index_second_constant(self):
        m = self.get_model()
        m.z = Var(m.I, m.I, domain=Integers)

        e = m.z[m.x, 3]

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        z = {}
        for i in [6, 7, 8]:
            self.assertIn(id(m.z[i, 3]), visitor.var_map)
            z[i, 3] = visitor.var_map[id(m.z[i, 3])]
            self.assertIs(visitor.pyomo_to_docplex[m.z[i, 3]], z[i, 3])
        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)

        self.assertTrue(
            expr[1].equals(
                cp.element([z[i, 3] for i in [6, 7, 8]], 0 + 1 * (x - 6) // 1)
            )
        )

    def test_indirection_multi_index_first_constant(self):
        m = self.get_model()
        m.z = Var(m.I, m.I, domain=Integers)

        e = m.z[3, m.x]

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        z = {}
        for i in [6, 7, 8]:
            self.assertIn(id(m.z[3, i]), visitor.var_map)
            z[3, i] = visitor.var_map[id(m.z[3, i])]
            self.assertIs(visitor.pyomo_to_docplex[m.z[3, i]], z[3, i])

        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)

        self.assertTrue(
            expr[1].equals(
                cp.element([z[3, i] for i in [6, 7, 8]], 0 + 1 * (x - 6) // 1)
            )
        )

    def test_indirection_multi_index_neither_constant_same_var(self):
        m = self.get_model()
        m.z = Var(m.I, m.I, domain=Integers)

        e = m.z[m.x, m.x]

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        z = {}
        for i in [6, 7, 8]:
            for j in [6, 7, 8]:
                self.assertIn(id(m.z[i, j]), visitor.var_map)
                z[i, j] = visitor.var_map[id(m.z[i, j])]
                self.assertIs(visitor.pyomo_to_docplex[m.z[i, j]], z[i, j])

        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)

        self.assertTrue(
            expr[1].equals(
                cp.element(
                    [z[i, j] for i in [6, 7, 8] for j in [6, 7, 8]],
                    0 + 1 * (x - 6) // 1 + 3 * (x - 6) // 1,
                )
            )
        )

    def test_indirection_multi_index_neither_constant_diff_vars(self):
        m = self.get_model()
        m.z = Var(m.I, m.I, domain=Integers)
        m.y = Var(within=[1, 3, 5])

        e = m.z[m.x, m.y]

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        z = {}
        for i in [6, 7, 8]:
            for j in [1, 3, 5]:
                self.assertIn(id(m.z[i, j]), visitor.var_map)
                z[i, j] = visitor.var_map[id(m.z[i, j])]
                self.assertIs(visitor.pyomo_to_docplex[m.z[i, j]], z[i, j])

        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)

        self.assertIn(id(m.y), visitor.var_map)
        y = visitor.var_map[id(m.y)]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)

        self.assertTrue(
            expr[1].equals(
                cp.element(
                    [z[i, j] for i in [6, 7, 8] for j in [1, 3, 5]],
                    0 + 1 * (x - 6) // 1 + 3 * (y - 1) // 2,
                )
            )
        )

    def test_indirection_expression_index(self):
        m = self.get_model()
        m.a.domain = Integers
        m.y = Var(within=[1, 3, 5])

        e = m.a[m.x - m.y]

        visitor = self.get_visitor()
        expr = visitor.walk_expression((e, e, 0))

        a = {}
        for i in range(1, 8):
            self.assertIn(id(m.a[i]), visitor.var_map)
            a[i] = visitor.var_map[id(m.a[i])]
            self.assertIs(visitor.pyomo_to_docplex[m.a[i]], a[i])

        self.assertIn(id(m.x), visitor.var_map)
        x = visitor.var_map[id(m.x)]
        self.assertIs(visitor.pyomo_to_docplex[m.x], x)
        self.assertIn(id(m.y), visitor.var_map)
        y = visitor.var_map[id(m.y)]
        self.assertIs(visitor.pyomo_to_docplex[m.y], y)

        self.assertTrue(
            expr[1].equals(
                cp.element([a[i] for i in range(1, 8)], 0 + 1 * (x + -1 * y - 1) // 1)
            )
        )

    def test_indirection_fails_with_non_finite_index_domain(self):
        m = self.get_model()
        m.a.domain = Integers
        # release the bounds
        m.x.setlb(None)
        m.x.setub(None)
        m.c = Constraint(expr=m.a[m.x] >= 0)

        visitor = self.get_visitor()
        with self.assertRaisesRegex(
            ValueError,
            r"Variable indirection 'a\[x\]' contains argument 'x', "
            "which is not restricted to a finite discrete domain",
        ):
            expr = visitor.walk_expression((m.c.body, m.c, 0))

    def test_indirection_invalid_index_domain(self):
        m = self.get_model()
        m.a.domain = Integers
        m.a.bounds = (6, 8)
        m.y = Var(within=Integers, bounds=(0, 10))

        e = m.a[m.y]

        visitor = self.get_visitor()
        with self.assertRaisesRegex(
            ValueError,
            r"Variable indirection 'a\[y\]' permits an index '0' "
            "that is not a valid key.",
        ):
            expr = visitor.walk_expression((e, e, 0))

    def test_infinite_domain_var(self):
        m = ConcreteModel()
        m.Evens = RangeSet(ranges=(NumericRange(0, None, 2), NumericRange(0, None, -2)))
        m.x = Var(domain=m.Evens)
        e = m.x**2

        visitor = self.get_visitor()
        with self.assertRaisesRegex(
            ValueError,
            "The LogicalToDoCplex writer does not support "
            "infinite discrete domains. Cannot "
            "write Var 'x' with domain 'Evens'",
        ):
            expr = visitor.walk_expression((e, e, 0))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for model preprocessing
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.environ import AbstractModel, Set, Param, Var, Objective


class TestPreprocess(unittest.TestCase):
    def Xtest_label1(self):
        model = AbstractModel()
        model.A = Set(initialize=[1, 2, 3])
        model.B = Param(model.A, initialize={1: 100, 2: 200, 3: 300})
        model.x = Var(model.A)
        model.y = Var(model.A)
        instance = model.create_instance()
        instance.preprocess()
        self.assertEqual(instance.num_used_variables(), 0)

    def Xtest_label2(self):
        model = AbstractModel()
        model.A = Set(initialize=[1, 2, 3])
        model.B = Param(model.A, initialize={1: 100, 2: 200, 3: 300})
        model.x = Var(model.A)
        model.y = Var(model.A)
        model.obj = Objective(rule=lambda inst: inst.x[1])
        instance = model.create_instance()
        instance.preprocess()
        self.assertEqual(instance.num_used_variables(), 1)
        self.assertEqual(instance.x[1].label, "x(1)")
        self.assertEqual(instance.x[2].label, "x(2)")
        self.assertEqual(instance.y[1].label, "y(1)")


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import bisect
from pyomo.common.numeric_types import native_numeric_types
from pyomo.common.timing import ConstructionTimer
from pyomo.core.base.set import SortedScalarSet
from pyomo.core.base.component import ModelComponentFactory

logger = logging.getLogger('pyomo.dae')


@ModelComponentFactory.register(
    "A bounded continuous numerical range optionally containing"
    " discrete points of interest."
)
class ContinuousSet(SortedScalarSet):
    """Represents a bounded continuous domain

    Minimally, this set must contain two numeric values defining the
    bounds of a continuous range. Discrete points of interest may
    be added to the continuous set. A continuous set is one
    dimensional and may only contain numerical values.

    Parameters
    ----------
    initialize : `list`
        Default discretization points to be included

    bounds : `tuple`
        The bounding points for the continuous domain. The bounds will
        be included as discrete points in the :py:class:`ContinuousSet`
        and will be used to bound the points added to the
        :py:class:`ContinuousSet` through the 'initialize' argument,
        a data file, or the add() method

    Attributes
    ----------
    _changed : `boolean`
        This keeps track of whether or not the ContinuousSet was changed
        during discretization. If the user specifies all of the needed
        discretization points before the discretization then there is no
        need to go back through the model and reconstruct things indexed
        by the :py:class:`ContinuousSet`

    _fe : `list`
        This is a sorted list of the finite element points in the
        :py:class:`ContinuousSet`. i.e. this list contains all the
        discrete points in the :py:class:`ContinuousSet` that are not
        collocation points. Points that are both finite element points
        and collocation points will be included in this list.

    _discretization_info : `dict`
        This is a dictionary which contains information on the
        discretization transformation which has been applied to the
        :py:class:`ContinuousSet`.
    """

    def __init__(self, *args, **kwds):
        """Constructor"""
        if kwds.pop("filter", None) is not None:
            raise TypeError(
                "'filter' is not a valid keyword argument for ContinuousSet"
            )
        # if kwds.pop("within", None) is not None:
        #    raise TypeError("'within' is not a valid keyword argument for "
        #  ContinuousSet")
        kwds.setdefault('dimen', 1)
        if kwds["dimen"] != 1:
            raise TypeError("'dimen' is not a valid keyword argument for ContinuousSet")
        if kwds.pop("virtual", None) is not None:
            raise TypeError(
                "'virtual' is not a valid keyword argument for ContinuousSet"
            )
        if kwds.pop("validate", None) is not None:
            raise TypeError(
                "'validate' is not a valid keyword argument for ContinuousSet"
            )
        if len(args) != 0:
            raise TypeError("A ContinuousSet expects no arguments")

        kwds.setdefault('ctype', ContinuousSet)
        self._changed = False
        self._fe = []
        self._discretization_info = {}
        super(ContinuousSet, self).__init__(**kwds)

    def get_finite_elements(self):
        """Returns the finite element points

        If the :py:class:`ContinuousSet <pyomo.dae.ContinuousSet>` has been
        discretizaed using a collocation scheme, this method will return a
        list of the finite element discretization points but not the
        collocation points within each finite element. If the
        :py:class:`ContinuousSet <pyomo.dae.ContinuousSet>` has not been
        discretized or a finite difference discretization was used,
        this method returns a list of all the discretization points in the
        :py:class:`ContinuousSet <pyomo.dae.ContinuousSet>`.

        Returns
        -------
        `list` of `floats`
        """
        return self._fe

    def get_discretization_info(self):
        """Returns a `dict` with information on the discretization scheme
        that has been applied to the :py:class:`ContinuousSet`.

        Returns
        -------
        `dict`
        """
        return self._discretization_info

    def get_changed(self):
        """Returns flag indicating if the :py:class:`ContinuousSet` was
        changed during discretization

        Returns "True" if additional points were added to the
        :py:class:`ContinuousSet <pyomo.dae.ContinuousSet>` while applying a
        discretization scheme

        Returns
        -------
        `boolean`
        """
        return self._changed

    def set_changed(self, newvalue):
        """Sets the ``_changed`` flag to 'newvalue'

        Parameters
        ----------
        newvalue : `boolean`

        """
        # TODO: Check this if-statement
        if newvalue is not True and newvalue is not False:
            raise ValueError(
                "The _changed attribute on a ContinuousSet may "
                "only be set to True or False"
            )
        self._changed = newvalue

    def get_upper_element_boundary(self, point):
        """Returns the first finite element point that is greater or equal
        to 'point'

        Parameters
        ----------
        point : `float`

        Returns
        -------
        float
        """
        if point in self._fe:
            return point
        elif point > max(self._fe):
            logger.warning(
                "The point '%s' exceeds the upper bound "
                "of the ContinuousSet '%s'. Returning the upper bound"
                % (str(point), self.name)
            )
            return max(self._fe)
        else:
            for i in self._fe:
                # This works because the list _fe is always sorted
                if i > point:
                    return i

    def get_lower_element_boundary(self, point):
        """Returns the first finite element point that is less than or
        equal to 'point'

        Parameters
        ----------
        point : `float`

        Returns
        -------
        float
        """
        if point in self._fe:
            if 'scheme' in self._discretization_info:
                if self._discretization_info['scheme'] == 'LAGRANGE-RADAU':
                    # Because Radau Collocation has a collocation point on the
                    # upper finite element bound this if statement ensures that
                    # the desired finite element bound is returned
                    tmp = self._fe.index(point)
                    if tmp != 0:
                        return self._fe[tmp - 1]
            return point
        elif point < min(self._fe):
            logger.warning(
                "The point '%s' is less than the lower bound "
                "of the ContinuousSet '%s'. Returning the lower bound "
                % (str(point), self.name)
            )
            return min(self._fe)
        else:
            rev_fe = list(self._fe)
            rev_fe.reverse()
            for i in rev_fe:
                if i < point:
                    return i

    def construct(self, values=None):
        """Constructs a :py:class:`ContinuousSet` component"""
        if self._constructed:
            return
        timer = ConstructionTimer(self)
        super(ContinuousSet, self).construct(values)

        for val in self:
            if type(val) is tuple:
                raise ValueError("ContinuousSet cannot contain tuples")
            if val.__class__ not in native_numeric_types:
                raise ValueError("ContinuousSet can only contain numeric values")

        # TBD: If a user specifies bounds they will be added to the set
        # unless the user specified bounds have been overwritten during
        # OrderedScalarSet construction. This can lead to some unintuitive
        # behavior when the ContinuousSet is both initialized with values and
        # bounds are specified. The current implementation is consistent
        # with how 'Set' treats this situation.
        for bnd in self.domain.bounds():
            # Note: the base class constructor ensures that any declared
            # set members are already within the bounds.
            if bnd is not None and bnd not in self:
                self.add(bnd)

        if None in self.bounds():
            raise ValueError(
                "ContinuousSet '%s' must have at least two values"
                " indicating the range over which a differential "
                "equation is to be discretized" % self.name
            )

        if len(self) < 2:
            # (reachable if lb==ub)
            raise ValueError(
                "ContinuousSet '%s' must have at least two values"
                " indicating the range over which a differential "
                "equation is to be discretized" % self.name
            )
        self._fe = list(self)
        timer.report()

    def find_nearest_index(self, target, tolerance=None):
        """Returns the index of the nearest point in the
        :py:class:`ContinuousSet <pyomo.dae.ContinuousSet>`.

        If a tolerance is specified, the index will only be returned
        if the distance between the target and the closest point is
        less than or equal to that tolerance. If there is a tie for
        closest point, the index on the left is returned.

        Parameters
        ----------
        target : `float`
        tolerance : `float` or `None`

        Returns
        -------
        `float` or `None`
        """
        lo = 0
        hi = len(self)
        arr = list(self)
        i = bisect.bisect_right(arr, target, lo=lo, hi=hi)
        # i is the index at which target should be inserted if it is to be
        # right of any equal components.

        if i == lo:
            # target is less than every entry of the set
            nearest_index = i + 1
            delta = self.at(nearest_index) - target
        elif i == hi:
            # target is greater than or equal to every entry of the set
            nearest_index = i
            delta = target - self.at(nearest_index)
        else:
            # p_le <= target < p_g
            # delta_left = target - p_le
            # delta_right = p_g - target
            # delta = min(delta_left, delta_right)
            # Tie goes to the index on the left.
            delta, nearest_index = min(
                (abs(target - self.at(j)), j) for j in [i, i + 1]
            )

        if tolerance is not None:
            if delta > tolerance:
                return None
        return nearest_index



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import defaultdict
import itertools

import logging

from pyomo.environ import (
    TransformationFactory,
    Transformation,
    Var,
    Constraint,
    Objective,
    Any,
    value,
    BooleanVar,
    Connector,
    Expression,
    Suffix,
    Param,
    Set,
    SetOf,
    RangeSet,
    Block,
    ExternalFunction,
    SortComponents,
    LogicalConstraint,
)
from pyomo.common.autoslots import AutoSlots
from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.config import ConfigDict, ConfigValue, PositiveInt, InEnum
from pyomo.common.dependencies import attempt_import
from pyomo.common.dependencies import numpy as np, packaging
from pyomo.common.enums import IntEnum
from pyomo.common.modeling import unique_component_name
from pyomo.core.expr.numeric_expr import SumExpression
from pyomo.core.expr import identify_variables
from pyomo.core.expr import SumExpression
from pyomo.core.util import target_list
from pyomo.contrib.piecewise import PiecewiseLinearExpression, PiecewiseLinearFunction
from pyomo.gdp import Disjunct, Disjunction
from pyomo.network import Port
from pyomo.repn.quadratic import QuadraticRepnVisitor
from pyomo.repn.util import ExprType, OrderedVarRecorder


def _lt_importer():
    import lineartree
    import importlib.metadata

    # linear-tree through version 0.3.5 relies on a private method from
    # scikit-learn.  That method was removed in scikit-learn version
    # 1.7.0.  We will report linear-tree as "unavailable" if
    # scikit-learn is "too new."
    lt_ver = packaging.version.parse(importlib.metadata.version('linear-tree'))
    if lt_ver <= packaging.version.Version('0.3.5'):
        import sklearn

        skl_ver = packaging.version.parse(sklearn.__version__)
        if skl_ver >= packaging.version.Version('1.7.0'):
            raise ImportError(
                f"linear-tree<=0.3.5 (found {lt_ver}) is incompatible with "
                f"scikit-learn>=1.7.0 (found {skl_ver})"
            )
    return lineartree


lineartree, lineartree_available = attempt_import('lineartree', importer=_lt_importer)
sklearn_lm, sklearn_available = attempt_import('sklearn.linear_model')

logger = logging.getLogger(__name__)


class DomainPartitioningMethod(IntEnum):
    RANDOM_GRID = 1
    UNIFORM_GRID = 2
    LINEAR_MODEL_TREE_UNIFORM = 3
    LINEAR_MODEL_TREE_RANDOM = 4


class _NonlinearToPWLTransformationData(AutoSlots.Mixin):
    __slots__ = (
        'transformed_component',
        'src_component',
        'transformed_constraints',
        'transformed_objectives',
    )

    def __init__(self):
        self.transformed_component = ComponentMap()
        self.src_component = ComponentMap()
        self.transformed_constraints = defaultdict(ComponentSet)
        self.transformed_objectives = defaultdict(ComponentSet)


Block.register_private_data_initializer(_NonlinearToPWLTransformationData)


def _get_random_point_grid(bounds, n, func, config, seed=42):
    # Generate randomized grid of points
    linspaces = []
    np.random.seed(seed)
    for (lb, ub), is_integer in bounds:
        if not is_integer:
            linspaces.append(np.random.uniform(lb, ub, n))
        else:
            size = min(n, ub - lb + 1)
            linspaces.append(
                np.random.choice(range(lb, ub + 1), size=size, replace=False)
            )
    return list(itertools.product(*linspaces))


def _get_uniform_point_grid(bounds, n, func, config):
    # Generate non-randomized grid of points
    linspaces = []
    for (lb, ub), is_integer in bounds:
        if not is_integer:
            # Issues happen when exactly using the boundary
            nudge = (ub - lb) * 1e-4
            linspaces.append(np.linspace(lb + nudge, ub - nudge, n))
        else:
            size = min(n, ub - lb + 1)
            pts = np.linspace(lb, ub, size)
            linspaces.append(np.array([round(i) for i in pts]))
    return list(itertools.product(*linspaces))


def _get_points_lmt_random_sample(bounds, n, func, config, seed=42):
    points = _get_random_point_grid(bounds, n, func, config, seed=seed)
    return _get_points_lmt(points, bounds, func, config, seed)


def _get_points_lmt_uniform_sample(bounds, n, func, config, seed=42):
    points = _get_uniform_point_grid(bounds, n, func, config)
    return _get_points_lmt(points, bounds, func, config, seed)


def _get_points_lmt(points, bounds, func, config, seed):
    x_list = np.array(points)
    y_list = []

    for point in points:
        y_list.append(func(*point))
    max_depth = config.linear_tree_max_depth
    if max_depth is None:
        # Want the tree to grow with increasing points but not get too large.
        max_depth = max(4, int(np.log2(len(points) / 4)))
    regr = lineartree.LinearTreeRegressor(
        sklearn_lm.LinearRegression(),
        criterion='mse',
        max_bins=120,
        min_samples_leaf=4,
        max_depth=max_depth,
    )
    regr.fit(x_list, y_list)

    leaves, splits, thresholds = _parse_linear_tree_regressor(regr, bounds)

    bound_point_list = _generate_bound_points(leaves, bounds)
    return bound_point_list


_partition_method_dispatcher = {
    DomainPartitioningMethod.RANDOM_GRID: _get_random_point_grid,
    DomainPartitioningMethod.UNIFORM_GRID: _get_uniform_point_grid,
    DomainPartitioningMethod.LINEAR_MODEL_TREE_UNIFORM: _get_points_lmt_uniform_sample,
    DomainPartitioningMethod.LINEAR_MODEL_TREE_RANDOM: _get_points_lmt_random_sample,
}


def _get_pwl_function_approximation(func, config, bounds):
    """
    Get a piecewise-linear approximation of a function, given:

    func: function to approximate
    config: ConfigDict for transformation, specifying domain_partitioning_method,
       num_points, and max_depth (if using linear trees)
    bounds: list of tuples giving upper and lower bounds and a boolean indicating
       if the variable's domain is discrete or not, for each of func's arguments
    """
    method = config.domain_partitioning_method
    n = config.num_points
    points = _partition_method_dispatcher[method](bounds, n, func, config)

    # Don't confuse PiecewiseLinearFunction constructor...
    dim = len(points[0])
    if dim == 1:
        points = [pt[0] for pt in points]

    # After getting the points, construct PWLF using the
    # function-and-list-of-points constructor
    logger.debug(
        f"Constructing PWLF with {len(points)} points, each of which "
        f"are {dim}-dimensional"
    )
    return PiecewiseLinearFunction(points=points, function=func)


# Given a leaves dict (as generated by parse_tree) and a list of tuples
# representing variable bounds, generate the set of vertices separating each
# subset of the domain
def _generate_bound_points(leaves, bounds):
    bound_points = []
    for leaf in leaves.values():
        lower_corner_list = []
        upper_corner_list = []
        for var_bound in leaf['bounds'].values():
            lower_corner_list.append(var_bound[0])
            upper_corner_list.append(var_bound[1])

        for pt in [lower_corner_list, upper_corner_list]:
            for i in range(len(pt)):
                # clamp within bounds range
                pt[i] = max(pt[i], bounds[i][0][0])
                pt[i] = min(pt[i], bounds[i][0][1])

        if tuple(lower_corner_list) not in bound_points:
            bound_points.append(tuple(lower_corner_list))
        if tuple(upper_corner_list) not in bound_points:
            bound_points.append(tuple(upper_corner_list))

    # This process should have gotten every interior bound point. However, all
    # but two of the corners of the overall bounding box should have been
    # missed. Let's fix that now.
    for outer_corner in itertools.product(*[b[0] for b in bounds]):
        if outer_corner not in bound_points:
            bound_points.append(outer_corner)
    return bound_points


# Parse a LinearTreeRegressor and identify features such as bounds, slope, and
# intercept for leaves. Return some dicts.
def _parse_linear_tree_regressor(linear_tree_regressor, bounds):
    leaves = linear_tree_regressor.summary(only_leaves=True)
    splits = linear_tree_regressor.summary()

    for key, leaf in leaves.items():
        del splits[key]
        leaf['bounds'] = {}
        leaf['slope'] = list(leaf['models'].coef_)
        leaf['intercept'] = leaf['models'].intercept_

    L = np.array(list(leaves.keys()))
    features = np.arange(0, len(leaves[L[0]]['slope']))

    for node in splits.values():
        left_child_node = node['children'][0]  # find its left child
        right_child_node = node['children'][1]  # find its right child
        # create the list to save leaves
        node['left_leaves'], node['right_leaves'] = [], []
        if left_child_node in leaves:  # if left child is a leaf node
            node['left_leaves'].append(left_child_node)
        else:  # traverse its left node by calling function to find all the
            # leaves from its left node
            node['left_leaves'] = _find_leaves(splits, leaves, splits[left_child_node])
        if right_child_node in leaves:  # if right child is a leaf node
            node['right_leaves'].append(right_child_node)
        else:  # traverse its right node by calling function to find all the
            # leaves from its right node
            node['right_leaves'] = _find_leaves(
                splits, leaves, splits[right_child_node]
            )

    # For each feature in each leaf, initialize lower and upper bounds to None
    for th in features:
        for leaf in leaves:
            leaves[leaf]['bounds'][th] = [None, None]
    for split in splits:
        var = splits[split]['col']
        for leaf in splits[split]['left_leaves']:
            leaves[leaf]['bounds'][var][1] = splits[split]['th']

        for leaf in splits[split]['right_leaves']:
            leaves[leaf]['bounds'][var][0] = splits[split]['th']

    leaves_new = _reassign_none_bounds(leaves, bounds)
    splitting_thresholds = {}
    for split in splits:
        var = splits[split]['col']
        splitting_thresholds[var] = {}
    for split in splits:
        var = splits[split]['col']
        splitting_thresholds[var][split] = splits[split]['th']
    # Make sure every nested dictionary in the splitting_thresholds dictionary
    # is sorted by value
    for var in splitting_thresholds:
        splitting_thresholds[var] = dict(
            sorted(splitting_thresholds[var].items(), key=lambda x: x[1])
        )

    return leaves_new, splits, splitting_thresholds


# This doesn't catch all additively separable expressions--we really need a
# walker (as does gdp.partition_disjuncts)
def _additively_decompose_expr(input_expr, min_dimension):
    dimension = len(list(identify_variables(input_expr)))
    if input_expr.__class__ is not SumExpression or dimension < min_dimension:
        # This isn't separable or we don't want to separate it, so we just have
        # the one expression
        return [input_expr]
    # else, it was a SumExpression, and we will break it into the summands
    return list(input_expr.args)


# Populate the "None" bounds with the bounding box bounds for a leaves-dict-tree
# amalgamation.
def _reassign_none_bounds(leaves, input_bounds):
    L = np.array(list(leaves.keys()))
    features = np.arange(0, len(leaves[L[0]]['slope']))

    for l in L:
        for f in features:
            if leaves[l]['bounds'][f][0] == None:
                leaves[l]['bounds'][f][0] = input_bounds[f][0][0]
            if leaves[l]['bounds'][f][1] == None:
                leaves[l]['bounds'][f][1] = input_bounds[f][0][1]
    return leaves


def _find_leaves(splits, leaves, input_node):
    root_node = input_node
    leaves_list = []
    queue = [root_node]
    while queue:
        node = queue.pop()
        node_left = node['children'][0]
        node_right = node['children'][1]
        if node_left in leaves:
            leaves_list.append(node_left)
        else:
            queue.append(splits[node_left])
        if node_right in leaves:
            leaves_list.append(node_right)
        else:
            queue.append(splits[node_right])
    return leaves_list


@TransformationFactory.register(
    'contrib.piecewise.nonlinear_to_pwl',
    doc="Convert nonlinear constraints and objectives to piecewise-linear "
    "approximations.",
)
class NonlinearToPWL(Transformation):
    """
    Convert nonlinear constraints and objectives to piecewise-linear approximations.
    """

    CONFIG = ConfigDict('contrib.piecewise.nonlinear_to_pwl')
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets that will be approximated",
            doc="""
            This specifies the list of components to approximate. If None (default),
            the entire model is transformed. Note that if the transformation is
            done out of place, the list of targets should be attached to the model
            before it is cloned, and the list will specify the targets on the cloned
            instance.""",
        ),
    )
    CONFIG.declare(
        'num_points',
        ConfigValue(
            default=3,
            domain=PositiveInt,
            description="Number of breakpoints for each piecewise-linear approximation",
            doc="""
            Specifies the number of points in each function domain to triangulate in 
            order to construct the piecewise-linear approximation. Must be an integer
            greater than 1.""",
        ),
    )
    CONFIG.declare(
        'domain_partitioning_method',
        ConfigValue(
            default=DomainPartitioningMethod.UNIFORM_GRID,
            domain=InEnum(DomainPartitioningMethod),
            description="Method for sampling points that will partition function "
            "domains.",
            doc="""
            The method by which the points used to partition each function domain
            are selected. By default, the range of each variable is partitioned
            uniformly, however it is possible to sample randomly or to use the
            partitions from training a linear model tree based on either uniform
            or random samples of the ranges.""",
        ),
    )
    CONFIG.declare(
        'approximate_quadratic_constraints',
        ConfigValue(
            default=True,
            domain=bool,
            description="Whether or not to approximate quadratic constraints.",
            doc="""
            Whether or not to calculate piecewise-linear approximations for
            quadratic constraints. If True, the resulting approximation will be
            a mixed-integer linear program. If False, the resulting approximation
            will be a mixed-integer quadratic program.""",
        ),
    )
    CONFIG.declare(
        'approximate_quadratic_objectives',
        ConfigValue(
            default=True,
            domain=bool,
            description="Whether or not to approximate quadratic objectives.",
            doc="""
            Whether or not to calculate piecewise-linear approximations for
            quadratic objectives. If True, the resulting approximation will be
            a mixed-integer linear program. If False, the resulting approximation
            will be a mixed-integer quadratic program.""",
        ),
    )
    CONFIG.declare(
        'additively_decompose',
        ConfigValue(
            default=False,
            domain=bool,
            description="Whether or not to additively decompose constraint expressions "
            "and approximate the summands separately.",
            doc="""
            If False, each nonlinear constraint expression will be approximated by
            exactly one piecewise-linear function. If True, constraints will be 
            additively decomposed, and each of the resulting summands will be
            approximated by a separate piecewise-linear function.

            It is recommended to leave this False as long as no nonlinear constraint 
            involves more than about 5-6 variables. For constraints with higher-
            dimensional nonlinear functions, additive decomposition will improve
            the scalability of the approximation (since partitioning the domain is
            subject to the curse of dimensionality).""",
        ),
    )
    CONFIG.declare(
        'max_dimension',
        ConfigValue(
            default=5,
            domain=PositiveInt,
            description="The maximum dimension of functions that will be approximated.",
            doc="""
            Specifies the maximum dimension function the transformation should
            attempt to approximate. If a nonlinear function dimension exceeds
            'max_dimension' the transformation will log a warning and leave the
            expression as-is. For functions with dimension significantly above the
            default (5), it is likely that this transformation will stall
            triangulating the points in order to partition the function domain.""",
        ),
    )
    CONFIG.declare(
        'min_dimension_to_additively_decompose',
        ConfigValue(
            default=1,
            domain=PositiveInt,
            description="The minimum dimension of functions that will be additively "
            "decomposed.",
            doc="""
            Specifies the minimum dimension of a function that the transformation
            should attempt to additively decompose. If a nonlinear function dimension
            exceeds 'min_dimension_to_additively_decompose' the transformation will
            additively decompose. If a the dimension of an expression is less than
            the 'min_dimension_to_additively_decompose' then it will not be additively
            decomposed""",
        ),
    )
    CONFIG.declare(
        'linear_tree_max_depth',
        ConfigValue(
            default=None,
            domain=PositiveInt,
            description="Maximum depth for linear tree training, used if using a "
            "domain partitioning method based on linear model trees.",
            doc="""
            Only used if 'domain_partitioning_method' is LINEAR_MODEL_TREE_UNIFORM or
            LINEAR_MODEL_TREE_RANDOM: Specifies the maximum depth of the linear model
            trees trained to determine the points to be triangulated to form the 
            domain of the piecewise-linear approximations. If None (the default),
            the max depth will be given as max(4, ln(num_points / 4)).
            """,
        ),
    )

    def __init__(self):
        super(Transformation).__init__()
        self._handlers = {
            Constraint: self._transform_constraint,
            Objective: self._transform_objective,
            Var: False,
            BooleanVar: False,
            Connector: False,
            Expression: False,
            Suffix: False,
            Param: False,
            Set: False,
            SetOf: False,
            RangeSet: False,
            Disjunction: False,
            Disjunct: self._transform_block_components,
            Block: self._transform_block_components,
            ExternalFunction: False,
            Port: False,
            PiecewiseLinearFunction: False,
            LogicalConstraint: False,
        }
        self._transformation_blocks = {}
        self._transformation_block_set = ComponentSet()
        self._quadratic_repn_visitor = QuadraticRepnVisitor(
            subexpression_cache={},
            var_recorder=OrderedVarRecorder(var_map={}, var_order={}, sorter=None),
        )

    def _apply_to(self, instance, **kwds):
        try:
            self._apply_to_impl(instance, **kwds)
        finally:
            self._transformation_blocks.clear()
            self._transformation_block_set.clear()

    def _apply_to_impl(self, model, **kwds):
        config = self.CONFIG(kwds.pop('options', {}))
        config.set_value(kwds)

        targets = config.targets
        if targets is None:
            targets = (model,)

        for target in targets:
            if target.ctype is Block or target.ctype is Disjunct:
                self._transform_block_components(target, config)
            elif target.ctype is Constraint:
                self._transform_constraint(target, config)
            elif target.ctype is Objective:
                self._transform_objective(target, config)
            else:
                raise ValueError(
                    "Target '%s' is not a Block, Constraint, or Objective. It "
                    "is of type '%s' and cannot be transformed."
                    % (target.name, type(target))
                )

    def _get_transformation_block(self, parent):
        if parent in self._transformation_blocks:
            return self._transformation_blocks[parent]

        nm = unique_component_name(parent, '_pyomo_contrib_nonlinear_to_pwl')
        self._transformation_blocks[parent] = transBlock = Block()
        parent.add_component(nm, transBlock)
        self._transformation_block_set.add(transBlock)

        transBlock._pwl_cons = Constraint(Any)
        return transBlock

    def _transform_block_components(self, block, config):
        blocks = block.values() if block.is_indexed() else (block,)
        for b in blocks:
            for obj in b.component_objects(
                active=True, descend_into=False, sort=SortComponents.deterministic
            ):
                if obj in self._transformation_block_set:
                    # This is a Block we created--we know we don't need to look
                    # on it.
                    continue
                handler = self._handlers.get(obj.ctype, None)
                if not handler:
                    if handler is None:
                        raise RuntimeError(
                            "No transformation handler registered for modeling "
                            "components of type '%s'." % obj.ctype
                        )
                    continue
                handler(obj, config)

    def _transform_constraint(self, cons, config):
        trans_block = self._get_transformation_block(cons.parent_block())
        trans_data_dict = trans_block.private_data()
        src_data_dict = cons.parent_block().private_data()
        constraints = cons.values() if cons.is_indexed() else (cons,)
        for c in constraints:
            pw_approx, expr_type = self._approximate_expression(
                c.body, c, trans_block, config, config.approximate_quadratic_constraints
            )

            if pw_approx is None:
                # Didn't need approximated, nothing to do
                continue
            c.model().private_data().transformed_constraints[expr_type].add(c)

            idx = len(trans_block._pwl_cons)
            trans_block._pwl_cons[c.name, idx] = (c.lower, pw_approx, c.upper)
            new_cons = trans_block._pwl_cons[c.name, idx]
            trans_data_dict.src_component[new_cons] = c
            src_data_dict.transformed_component[c] = new_cons

            # deactivate original
            c.deactivate()

    def _transform_objective(self, objective, config):
        trans_block = self._get_transformation_block(objective.parent_block())
        trans_data_dict = trans_block.private_data()
        objectives = objective.values() if objective.is_indexed() else (objective,)
        src_data_dict = objective.parent_block().private_data()
        for obj in objectives:
            pw_approx, expr_type = self._approximate_expression(
                obj.expr,
                obj,
                trans_block,
                config,
                config.approximate_quadratic_objectives,
            )

            if pw_approx is None:
                # Didn't need approximated, nothing to do
                continue
            obj.model().private_data().transformed_objectives[expr_type].add(obj)

            new_obj = Objective(expr=pw_approx, sense=obj.sense)
            trans_block.add_component(
                unique_component_name(trans_block, obj.name), new_obj
            )
            trans_data_dict.src_component[new_obj] = obj
            src_data_dict.transformed_component[obj] = new_obj

            obj.deactivate()

    def _get_bounds_list(self, var_list, obj):
        bounds = []
        for v in var_list:
            if None in v.bounds:
                raise ValueError(
                    "Cannot automatically approximate constraints with unbounded "
                    "variables. Var '%s' appearing in component '%s' is missing "
                    "at least one bound" % (v.name, obj.name)
                )
            else:
                bounds.append((v.bounds, v.is_integer()))
        return bounds

    def _needs_approximating(self, expr, approximate_quadratic):
        repn = self._quadratic_repn_visitor.walk_expression(expr)
        if repn.nonlinear is None:
            if repn.quadratic is None:
                # Linear constraint. Always skip.
                return ExprType.LINEAR, False
            else:
                if not approximate_quadratic:
                    # Didn't need approximated, nothing to do
                    return ExprType.QUADRATIC, False
                return ExprType.QUADRATIC, True
        return ExprType.GENERAL, True

    def _approximate_expression(
        self, expr, obj, trans_block, config, approximate_quadratic
    ):
        expr_type, needs_approximating = self._needs_approximating(
            expr, approximate_quadratic
        )
        if not needs_approximating:
            return None, expr_type

        # Additively decompose expr and work on the pieces
        pwl_summands = []
        for k, subexpr in enumerate(
            _additively_decompose_expr(
                expr, config.min_dimension_to_additively_decompose
            )
            if config.additively_decompose
            else (expr,)
        ):
            # First check if this is a good idea
            expr_vars = list(identify_variables(subexpr, include_fixed=False))
            orig_values = ComponentMap((v, v.value) for v in expr_vars)

            dim = len(expr_vars)
            if dim > config.max_dimension:
                raise ValueError(
                    "Not approximating expression for component '%s' as "
                    "it exceeds the maximum dimension of %s. Try increasing "
                    "'max_dimension' or additively separating the expression."
                    % (obj.name, config.max_dimension)
                )
                pwl_summands.append(subexpr)
                continue
            elif not self._needs_approximating(subexpr, approximate_quadratic)[1]:
                pwl_summands.append(subexpr)
                continue
            # else we approximate subexpr

            def eval_expr(*args):
                for i, v in enumerate(expr_vars):
                    v.value = args[i]
                return value(subexpr)

            pwlf = _get_pwl_function_approximation(
                eval_expr, config, self._get_bounds_list(expr_vars, obj)
            )
            name = unique_component_name(
                trans_block, obj.getname(fully_qualified=False)
            )
            trans_block.add_component(f"_pwle_{name}_{k}", pwlf)
            # NOTE: We are *not* using += because it will hit the NamedExpression
            # implementation of iadd and dereference the ExpressionData holding
            # the PiecewiseLinearExpression that we later transform my remapping
            # it to a Var...
            pwl_summands.append(pwlf(*expr_vars))

            # restore var values
            for v, val in orig_values.items():
                v.value = val

        return sum(pwl_summands), expr_type

    def get_src_component(self, cons):
        data = cons.parent_block().private_data().src_component
        if cons in data:
            return data[cons]
        else:
            raise ValueError(
                "It does not appear that '%s' is a transformed Constraint "
                "created by the 'nonlinear_to_pwl' transformation." % cons.name
            )

    def get_transformed_component(self, cons):
        data = cons.parent_block().private_data().transformed_component
        if cons in data:
            return data[cons]
        else:
            raise ValueError(
                "It does not appear that '%s' is a Constraint that was "
                "transformed by the 'nonlinear_to_pwl' transformation." % cons.name
            )

    def get_transformed_nonlinear_constraints(self, model):
        """
        Given a model that has been transformed with contrib.piecewise.nonlinear_to_pwl,
        return the list of general (not quadratic) nonlinear Constraints that were
        approximated with PiecewiseLinearFunctions
        """
        return model.private_data().transformed_constraints[ExprType.GENERAL]

    def get_transformed_quadratic_constraints(self, model):
        """
        Given a model that has been transformed with contrib.piecewise.nonlinear_to_pwl,
        return the list of quadratic Constraints that were approximated with
        PiecewiseLinearFunctions
        """
        return model.private_data().transformed_constraints[ExprType.QUADRATIC]

    def get_transformed_nonlinear_objectives(self, model):
        """
        Given a model that has been transformed with contrib.piecewise.nonlinear_to_pwl,
        return the list of general (not quadratic) nonlinear Constraints that were
        approximated with PiecewiseLinearFunctions
        """
        return model.private_data().transformed_objectives[ExprType.GENERAL]

    def get_transformed_quadratic_objectives(self, model):
        """
        Given a model that has been transformed with contrib.piecewise.nonlinear_to_pwl,
        return the list of quadratic Constraints that were approximated with
        PiecewiseLinearFunctions
        """
        return model.private_data().transformed_objectives[ExprType.QUADRATIC]
